#
# Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#


import getpass
import hashlib
import logging
import re

from mysql.fabric import (
    config as _config,
    errors as _errors,
    persistence as _persistence,
    executor as _executor,
    command as _command,
)

_LOGGER = logging.getLogger(__name__)

FABRIC_REALM_XMLRPC = 'MySQL Fabric'

FABRIC_DEFAULT_PROTOCOL = 'xmlrpc'
FABRIC_PROTOCOL_DEFAULTS = {
    'protocol.xmlrpc': {
        'realm': FABRIC_REALM_XMLRPC,
    },
}

# This sequence is used when creating an dropping tables
_SQL_TABLES = ('users', 'roles', 'permissions', 'role_permissions',
               'user_roles')

_SQL_CREATE = dict()
_SQL_CREATE['users'] = """
CREATE TABLE `users` (
  `user_id` int unsigned NOT NULL AUTO_INCREMENT,
  `username` varchar(100) NOT NULL,
  `protocol` varchar(200) NOT NULL DEFAULT 'xmlrpc',
  `password` varchar(128) DEFAULT NULL,
  PRIMARY KEY (user_id),
  UNIQUE KEY (`username`, `protocol`)
) DEFAULT CHARSET=utf8
"""

_SQL_CREATE['roles'] = """
CREATE TABLE `roles` (
  `role_id` int unsigned NOT NULL AUTO_INCREMENT,
  `name` varchar(80) NOT NULL,
  `description` varchar(1000),
  PRIMARY KEY (role_id)
) DEFAULT CHARSET=utf8
"""

_SQL_CREATE['permissions'] = """
CREATE TABLE `permissions` (
  `permission_id` int unsigned NOT NULL AUTO_INCREMENT,
  `subsystem` varchar(60) NOT NULL,
  `component` varchar(60) DEFAULT NULL,
  `function` varchar(60) DEFAULT NULL,
  `description` varchar(1000),
  PRIMARY KEY (permission_id),
  UNIQUE INDEX (`subsystem`, `component`, `function`)
) DEFAULT CHARSET=utf8
"""

_SQL_CREATE['role_permissions'] = """
CREATE TABLE `role_permissions` (
  `role_id` int unsigned NOT NULL,
  `permission_id` int unsigned NOT NULL,
  PRIMARY KEY (`role_id`, `permission_id`)
) DEFAULT CHARSET=utf8
"""

_SQL_CREATE['user_roles'] = """
CREATE TABLE `user_roles` (
  `user_id` int unsigned NOT NULL,
  `role_id` int unsigned NOT NULL,
  PRIMARY KEY (`user_id`, `role_id`)
) DEFAULT CHARSET=utf8
"""

_SQL_CONSTRAINTS = [
    "ALTER TABLE user_roles ADD CONSTRAINT fk_user_roles_user_id "
    "FOREIGN KEY (user_id) REFERENCES users (user_id) ON DELETE CASCADE, "
    "ADD CONSTRAINT fk_user_roles_role_id FOREIGN KEY (role_id) "
    "REFERENCES roles (role_id) ON DELETE CASCADE",

    "ALTER TABLE role_permissions ADD CONSTRAINT fk_role_permissions_role_id "
    "FOREIGN KEY (role_id) REFERENCES roles (role_id), ADD CONSTRAINT "
    "fk_role_permissions_permission_id FOREIGN KEY (permission_id) "
    "REFERENCES permissions (permission_id)",
]

_USERS = [
    # user_id, username, protocol, password
    (1, 'admin', 'xmlrpc', None),
]

_PERMISSIONS = [
    (1, 'core', None, None, 'Full access to all core Fabric functionality'),
    (2, 'core', 'dump', None, 'Access to dump commands'),
    (3, 'core', 'user', None, 'User administration'),
    (4, 'core', 'role', None, 'Role administration'),
    (5, 'core', 'threat', None, 'Reporting to Fabric'),
]

_ROLES = [
    (1, 'superadmin', 'Role for Administrative users'),
    (2, 'useradmin', 'Role for users dealing with user administration'),
    (3, 'connector', 'Role for MySQL Connectors'),
]

_ROLE_PERMISSIONS = {
    1: (1,),  # superadmin permissions
    3: (2, 5),  # connector permissions
    2: (3, 4),  # useradmin permissions
}

_USER_ROLES = [
    (1, 1),  # admin in superadmin role
]

_SQL_FETCH_USER = """
SELECT username, protocol, password
FROM users WHERE username = %s AND protocol = %s
"""

_SQL_FETCH_USER_PERMISSIONS = """
SELECT p.subsystem, p.component, p.function
FROM users AS u LEFT JOIN user_roles AS ur USING (user_id)
    LEFT JOIN role_permissions AS rp USING (role_id)
    LEFT JOIN permissions AS p USING (permission_id)
WHERE
    u.username = %s AND protocol = %s
"""


class User(_persistence.Persistable):
    """Class defining a user connecting to a Fabric instance
    """

    def __init__(self, username, protocol, password_hash):
        self._username = username
        self._protocol = protocol
        self._password_hash = password_hash
        self._permissions = User.fetch_permissions(username, protocol)

    @property
    def username(self):
        return self._username

    @property
    def protocol(self):
        return self._protocol

    @property
    def password_hash(self):
        return self._password_hash

    @staticmethod
    def add_user(username, password, protocol, user_id=None,
                 config=None, persister=None):
        stmt = (
            "INSERT INTO users (user_id, username, password, protocol) "
            "VALUES (%(user_id)s, %(username)s, %(password)s, %(protocol)s)"
        )

        hashed_password = _hash_password(username, password, protocol, config)

        options = {
            "raw": False,
            "fetch": False,
            "params": {
                'user_id': user_id,
                'username': username,
                'protocol': protocol,
                'password': hashed_password,
            },
        }
        cur = persister.exec_stmt(stmt, options)
        return cur.lastrowid

    @staticmethod
    def add_permission(subsystem, component=None, function=None,
                       description=None, permission_id=None, persister=None):
        stmt = (
            "INSERT INTO permissions (permission_id, subsystem, "
            "component, function, description) "
            "VALUES (%(permission_id)s, %(subsystem)s, %(component)s, "
            "%(function)s, %(description)s)"
        )

        options = {
            "raw": False,
            "fetch": False,
            "params": {
                'permission_id': permission_id,
                'subsystem': subsystem,
                'function': function,
                'component': component,
                'description': description,
            },
        }
        persister.exec_stmt(stmt, options)

    @staticmethod
    def add_role(name, description, role_id=None, persister=None):
        stmt = (
            "INSERT INTO roles (role_id, name, description) "
            "VALUES (%(role_id)s, %(name)s, %(description)s)"
        )

        options = {
            "raw": False,
            "fetch": False,
            "params": {
                'role_id': role_id,
                'name': name,
                'description': description,
            },
        }
        persister.exec_stmt(stmt, options)

    @staticmethod
    def add_role_permission(role_id, permission_id, persister=None):
        stmt = (
            "INSERT INTO role_permissions (role_id, permission_id) "
            "VALUES (%(role_id)s, %(permission_id)s)"
        )

        options = {
            "raw": False,
            "fetch": False,
            "params": {
                'role_id': role_id,
                'permission_id': permission_id,
            },
        }
        persister.exec_stmt(stmt, options)

    @staticmethod
    def add_user_role(user_id, role_id, persister=None):
        stmt = (
            "INSERT INTO user_roles (user_id, role_id) "
            "VALUES (%(user_id)s, %(role_id)s)"
        )

        options = {
            "raw": False,
            "fetch": False,
            "params": {
                'role_id': role_id,
                'user_id': user_id,
            },
        }
        persister.exec_stmt(stmt, options)

    @staticmethod
    def delete_user(user_id=None, username=None, protocol=None, persister=None):
        """Delete a Fabric user"""
        if (user_id and username) or (not user_id and not username):
            raise AttributeError("Use user_id or username, not both")

        if username and not protocol:
            raise AttributeError(
                "protocol is required when username is provided")

        options = {
            "raw": False,
            "fetch": False,
            "params": {},
        }
        if user_id:
            options['params']['user_id'] = user_id
            sql = "DELETE FROM users WHERE user_id = %(user_id)s"
        else:
            options['params']['username'] = username
            options['params']['protocol'] = protocol
            sql = ("DELETE FROM users WHERE username = %(username)s "
                   "AND protocol = %(protocol)s")

        persister.exec_stmt(sql, options)

    @staticmethod
    def create(persister=None, config=None):
        """Create table for FabricUser

        :param persister: A valid handle to the state store.
        """
        for table_name in _SQL_TABLES:
            persister.exec_stmt(_SQL_CREATE[table_name])

        for statement in _SQL_CONSTRAINTS:
            persister.exec_stmt(statement)

        for user_id, username, protocol, password in _USERS:
            User.add_user(username, password, protocol,
                          user_id=user_id,
                          persister=persister,
                          config=config)

        for permission_id, subsystem, component, function, description in \
                _PERMISSIONS:
            User.add_permission(
                subsystem, component, function, description,
                permission_id=permission_id, persister=persister)

        for role_id, name, description in _ROLES:
            User.add_role(name, description, role_id=role_id,
                          persister=persister)

        for role_id, permissions in _ROLE_PERMISSIONS.items():
            for permission_id in permissions:
                User.add_role_permission(
                    role_id, permission_id, persister=persister)

        for user_id, role_id in _USER_ROLES:
            User.add_user_role(user_id, role_id,
                               persister=persister)

    @staticmethod
    def fetch_user(username, protocol, realm, persister=None):
        """Fetch the information about a user using particular protocol

        :param username: Username being queried.
        :param protocol: Protocol with which the user connect.
        :param realm: Realm which the user belongs to.
        :param persister: A valid handle to the state store.
        :return: Returns user information including password hash.
        """
        if not protocol:
            protocol = FABRIC_DEFAULT_PROTOCOL

        options = {
            "raw": False,
            "fetch": False,
            "params": [username, protocol],
            "columns": True,
        }

        cur = persister.exec_stmt(_SQL_FETCH_USER, options)
        row = cur.fetchone()
        if row:
            return User(username=row.username,
                        protocol=row.protocol,
                        password_hash=row.password)

        return None

    @staticmethod
    def fetch_permissions(username, protocol, persister=None):
        """Fetch permissions for given user and protocol

        :param username: Username being queried.
        :param protocol: Protocol in which the user lives
        :param persister: A valid handle to the state store.
        :return: Returns user permissions as a list of tuples
        """
        if not protocol:
            protocol = FABRIC_DEFAULT_PROTOCOL

        options = {
            "raw": False,
            "fetch": False,
            "params": (username, protocol),
        }

        cur = persister.exec_stmt(_SQL_FETCH_USER_PERMISSIONS, options)
        rows = cur.fetchall()
        if rows:
            return rows

        return None

    def has_permission(self, subsystem, component=None, function=None):
        """Check if user has permission"""
        needles = [
            (subsystem, None, None),  # full access to subsystem
            (subsystem, component, None),  # access to all component's functions
            (subsystem, component, function)  # access to function
        ]

        for needle in needles:
            if needle in self._permissions:
                return True

        return False


def _hash_password(username, password, protocol, config, realm=None):
    """Hash password base on protocol

    :raises errors.CredentialError: if protocol is not implemented
    """
    if not password:
        return
    if not realm and config:
        section = 'protocol.' + protocol
        realm = config.get(section, 'realm',
                           vars=FABRIC_PROTOCOL_DEFAULTS[section])

    if protocol == 'xmlrpc':
        return hashlib.md5('{user}:{realm}:{secret}'.format(
            user=username, realm=FABRIC_REALM_XMLRPC,
            secret=password)).hexdigest()

    raise _errors.CredentialError(
        "Password hasing for protocol '{0}' is not implemented.".format(
            protocol
        ))


def get_user(username, protocol, persister=None):
    """Get user information from data store

    :param username: User to query for.
    :param protocol: Protocol used to connect.
    :param persister: A valid handle to the state store.
    :return: namedtuple with password, username, protocol and user_id
    :rtype: namedtuple
    """
    if not persister:
        persister = _persistence.current_persister()

    options = {
        "raw": False,
        "fetch": False,
        "params": (username, protocol),
        "columns": True,
    }

    cur = persister.exec_stmt(
        "SELECT password, username, protocol, user_id "
        "FROM users WHERE username = %s "
        "AND protocol = %s", options)
    row = cur.fetchone()
    if row:
        return row
    return None


def _get_password(prompt=None):
    """Get password from using STDIN

    :param prompt: String describing what input is required.
    :returns: Password as string.
    :rtype: str
    :raises CredentialError: if passwords did not match
    """
    if not prompt:
        prompt = 'Password: '
    password = getpass.getpass(prompt).strip()
    repeat = getpass.getpass('Repeat Password: ').strip()
    if password != repeat:
        raise _errors.CredentialError("Passwords did not match!")
    return password


def check_initial_setup(config, persister, check_only=False):
    """Check if admin user has password and if not sets it

    :param persister: A valid handle to the state store.
    """

    # Fetch which protocols have no passwords set for user 'admin'
    protocols = []
    for key in FABRIC_PROTOCOL_DEFAULTS.keys():
        if key.startswith('protocol.'):
            tmp = key.split('.', 2)[1]
            if tmp not in protocols:
                user = get_user('admin', tmp, persister)
                if not user.password:
                    protocols.append(tmp)

    # Try setting password for 'admin' user from configuration file
    for protocol in protocols:
        section = 'protocol.' + protocol
        try:
            username = config.get(section, 'user')
        except _config.NoOptionError:
            username = 'admin'

        # If there is no password, we have to ask for one.
        try:
            password = config.get(section, 'password').strip()

            # No password, so we have to ask for one
            if password == '':
                break
        except _config.NoOptionError:
            # No password, so we have to ask for one
            break

        persister.begin()
        try:
            if username != 'admin':
                _LOGGER.info("Adding user %s/%s", username, protocol)
                user_id = User.add_user(username, password, protocol,
                                        persister=persister)
            else:
                _LOGGER.info("Initial password for %s/%s set", username,
                             protocol)
                _change_password(username, password, protocol, config,
                                 persister)
        except _errors.CredentialError as error:
            print "Setting password cancelled."
            _LOGGER.debug(str(error))
            print error
            persister.rollback()
        else:
            persister.commit()
            msg = (
                "Password set for {user}/{protocol} from configuration file."
            ).format(user=username, protocol=protocol)
            print msg
            _LOGGER.info(msg)

        if username != 'admin':
            print "Note: {user}/{protocol} has no roles set!".format(
                user=username, protocol=protocol
            )
        else:
            # No need to ask for password later for this protocol
            protocols.remove(protocol)

    if not protocols:
        # Passwords are set
        return

    if check_only and protocols:
        print (
            "\nPassword for admin user is empty. Please run\n\n"
            "  shell> mysqlfabric user password admin\n\n"
            "Make sure the password is empty or commented in section\n"
            "[protocol.xmlrpc] of the configuration file before executing the\n"
            "above command."
        )
        raise _errors.CredentialError("Check empty admin password failed")

    print "Finishing initial setup"
    print "======================="
    print "Password for admin user is not yet set."
    password = None

    while True:
        password = _get_password("Password for {user}/{protocol}: ".format(
            user='admin', protocol='xmlrpc'))
        if password:
            break

    if password:
        for protocol in protocols:
            try:
                _change_password('admin', password, protocol, config,
                                 persister)
            except _errors.CredentialError as error:
                print "Setting password cancelled."
                _LOGGER.debug(str(error))
                print error
            else:
                print "Password set."
    else:
        # Making sure password is set and there is an error message
        raise _errors.CredentialError(
            "Password for admin can not be empty. Use `user password` command.")

    persister.commit()


def _configure_connections(config):
    """Configure information on database connection and remote servers.

    :param config: Configuration read from configuration file.
    """
    # Configure the number of concurrent executors.
    executor = _executor.Executor()
    executor.set_number_executors(1)

    # Fetch options to configure the state store.
    address = config.get('storage', 'address')

    try:
        host, port = address.split(':')
        port = int(port)
    except ValueError:
        host = address
        port = 3306

    user = config.get('storage', 'user')
    database = config.get('storage', 'database')

    try:
        password = config.get('storage', 'password')
    except _config.NoOptionError:
        password = _get_password("Storage password: ")

    try:
        connection_timeout = config.get("storage", "connection_timeout")
        connection_timeout = float(connection_timeout)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        connection_timeout = None

    try:
        connection_attempts = config.get("storage", "connection_attempts")
        connection_attempts = int(connection_attempts)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        connection_attempts = None

    try:
        connection_delay = config.get("storage", "connection_delay")
        connection_delay = int(connection_delay)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        connection_delay = None

    try:
        auth_plugin = config.get("storage", "auth_plugin")
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        auth_plugin = None

    # Define state store configuration.
    _persistence.init(
        host=host, port=port, user=user, password=password, database=database,
        connection_timeout=connection_timeout,
        connection_attempts=connection_attempts,
        connection_delay=connection_delay,
        auth_plugin=auth_plugin
    )


def confirm(message, default='y'):
    """Prompt confirming using Yes or No

    :param message: Message to be shown.
    :param default: The default answer, either 'y' or 'n'.
    :returns: True if answer was 'y', False if 'n'.
    :rtype: bool
    :raises ValueError: if invalid choice has been given.
    """
    if not isinstance(default, str):
        raise AttributeError("default argument should be a string")
    if default:
        default = default[0].lower()

    yes = 'Y' if default == 'y' else 'y'
    no = 'N' if default == 'n' else 'n'

    msg = message + " [{0}/{1}] ".format(yes, no)
    try:
        choice = raw_input(msg)[0].lower()
    except IndexError:
        choice = default

    if choice == 'y':
        return True
    elif choice == 'n':
        return False

    raise ValueError("Invalid choice")


def _role_listing(selected=None, marker=None, persister=None):
    """Display list of roles

    :param selected: Sequence of selected roles.
    :param marker: Marker used for showing selected roles.

    The selected argument should be a list of role IDs. The marker will be
    shown in front of the role ID when it is in the selected sequence.
    """
    if not persister:
        persister = _persistence.current_persister()

    if not selected:
        selected = []

    if marker:
        marker = marker[0]
    else:
        marker = 'X'

    options = {
        "raw": False,
        "fetch": False,
        "columns": True,
    }

    roleperms = (
        "SELECT r.role_id, r.name, r.description AS role_desc, "
        "p.permission_id, p.description AS perm_desc, p.subsystem, "
        "p.component, p.function "
        "FROM roles AS r LEFT JOIN role_permissions USING (role_id) "
        "LEFT JOIN permissions AS p USING (permission_id) ORDER BY r.role_id"
    )
    cur = persister.exec_stmt(roleperms, options)
    roles = {}
    max_name_len = 0
    max_rowid = 0
    for row in cur:
        if len(row.name) > max_name_len:
            max_name_len = len(row.name)
        if row.role_id > max_rowid:
            max_rowid = row.role_id

        if row.role_id not in roles:
            roles[row.role_id] = [row.name, row.role_desc, [row.perm_desc]]
        else:
            roles[row.role_id][2].append(row.perm_desc)

    # Minimum sizes
    max_rowid_len = max(2, len(str(max_rowid)))
    max_name_len = max(9, max_name_len)

    sel = ' '

    role_fmt = ("{{sel}} {{role_id:{idlen}}}  {{name:{namelen}}}"
                "  {{desc}}".format(idlen=max_rowid_len, namelen=max_name_len))

    label_desc = "Description and Permissions"
    header = role_fmt.format(role_id="ID", name="Role Name", desc=label_desc,
                             sel=sel)
    print header
    print role_fmt.format(role_id='-' * max_rowid_len, name='-' * max_name_len,
                          desc='-' * len(label_desc), sel=sel)

    fmt_perm = '{0}+ {{perm}}'.format(
        (2 + max_rowid_len + 2 + max_name_len + 2) * ' ')

    for role_id in sorted(roles.keys()):
        if role_id in selected:
            sel = marker
        else:
            sel = ' '
        name, role_desc, permissions = roles[role_id]
        print role_fmt.format(role_id=role_id, name=name, desc=role_desc,
                              sel=sel)
        for perm in permissions:
            print fmt_perm.format(perm=perm)


def _role_selection(message=None, choices=None, persister=None):
    """Offers user to select roles on the console

    :param persister: A valid handle to the state store.
    :param message: Message shown just before prompt.
    :param choices: Do not ask, just process choices (string or sequence).
    :return: List of role IDs or role names.
    :rtype: list
    :raises errors.CredentialError: if invalid role was given
    """
    if not persister:
        persister = _persistence.current_persister()

    if not choices:
        if not message:
            message = "\nEnter comma separated list of role IDs or names: "

        choices = raw_input(message)
        if not choices.strip():
            return []

    if isinstance(choices, str):
        choices = choices.split(',')

    options = {
        "raw": False,
        "fetch": False,
        "columns": True,
    }
    cur = persister.exec_stmt("SELECT role_id, name FROM roles", options)

    valid_role_ids = []
    roles = {}
    for row in cur:
        roles[row.name] = row.role_id
        if str(row.role_id) not in valid_role_ids:
            valid_role_ids.extend([str(row.role_id), row.name])

    try:
        if not all(rid.strip() in valid_role_ids for rid in choices):
            raise ValueError
    except ValueError:
        raise _errors.CredentialError("Found invalid role.")

    # Only return role IDs
    result = []
    for rid in choices:
        try:
            result.append(int(rid))
        except ValueError:
            # We got name
            result.append(roles[rid.strip()])

    return result


def _change_password(username, password, protocol, config, persister):
    """Change password of a Fabric user

    :param username: Username of Fabric user.
    :param password: Password to which we change.
    :param protocol: Protocol for this user.
    :param config: Fabric configuration.
    :param persister: A valid handle to the state store.

    :raise _errors.CredentialError: if any error occurs while updating data
    """
    try:
        persister.begin()
        options = {
            "raw": False,
            "fetch": False,
            "params": (),
            "columns": True,
        }
        update = ("UPDATE users SET password = %s WHERE username = %s"
                  " AND protocol = %s")
        hashed = _hash_password(username, password, protocol,
                                config)
        options['params'] = (hashed, username, protocol)
        persister.exec_stmt(update, options)
    except Exception as error:
        # We rollback and re-raise
        persister.rollback()
        raise _errors.CredentialError("Error updating password: {0}".format(
            str(error)
        ))
    persister.commit()


def validate_username(username, allow_empty=False):
    """Validates a username

    :param username: The username we need to check.
    :param allow_empty: Whether to allow empty username.
    :returns: Stripped username or None when username was empty
    :rtype: str
    :raises ValueError: if username is not acceptable.
    """
    username = username.strip()
    if not username and allow_empty:
        return None
    valid = r"[a-zA-Z0-9_@.-]"
    match = re.match(r"^" + valid + "+$", username)
    if not match:
        raise _errors.CredentialError("Invalid username (was not {0})".format(
            valid
        ))

    return username


def validate_protocol(protocol, allow_empty=False):
    """Validates a protocol

    :param protocol: The protocol we need to check.
    :param allow_empty: Whether to allow empty protocol.
    :returns: Stripped protocol or None when protocol was empty
    :rtype: str
    :raises ValueError: if protocol is not acceptable.
    """
    protocol = protocol.strip()
    if not protocol and allow_empty:
        return None
    valid = r"[a-zA-Z0-9_@.-]"
    match = re.match(r"^" + valid + "+$", protocol)
    if not match:
        raise _errors.CredentialError("Invalid protocol (was not {0})".format(
            valid
        ))

    return protocol


class UserCommand(_command.Command):
    """Base class for all user commands"""

    group_name = 'user'
    command_name = ''
    description = ''
    persister = None

    def dispatch(self, *args):
        """Setup Fabric Storage System.
        """
        # Configure connections.
        _configure_connections(self.config)
        _persistence.init_thread()
        self.persister = _persistence.current_persister()
        self.execute(*args)

    def execute(self, *args):
        raise NotImplementedError

    def _ask_credentials(self, username, ask_password=True, title=None):
        """Ask for credentials

        :param username: Username will not be asked if given.
        :param ask_password: Whether to ask for password or not.
        :param title: The title to show.
        :return: Tuple containing username, password and protocol
        :rtype: tuple
        """
        if not title:
            title = self.description

        print title
        print "=" * len(title)

        if username:
            print "Username: {0}".format(username)
        else:
            username = validate_username(raw_input("Enter username: "))

        if self.options.protocol:
            protocol = self.options.protocol
            print "Protocol: {0}".format(protocol)
        else:
            prompt = "Protocol (default {0}): ".format(FABRIC_DEFAULT_PROTOCOL)
            protocol = validate_protocol(raw_input(prompt),
                                         allow_empty=True)
            if not protocol:
                protocol = FABRIC_DEFAULT_PROTOCOL

        if ask_password:
            password = _get_password()
        else:
            password = None

        return username, password, protocol


class UserAdd(UserCommand):

    """Add a new Fabric user.

    * protocol: Protocol of the user (for example 'xmlrpc')
    * roles: Comma separated list of roles, IDs or names (see `role list`)

    """

    command_name = 'add'
    description = 'Add a new Fabric user'

    def execute(self, username, protocol=None, roles=None):
        """Add a new Fabric user"""
        username, password, protocol = self._ask_credentials(
            username, ask_password=False)

        # Check if user exists
        if get_user(username, protocol, persister=self.persister):
            raise _errors.CredentialError(
                "User {user}/{protocol} already exists".format(
                    user=username, protocol=protocol))

        password = _get_password('Password: ')

        if not password:
            raise _errors.CredentialError("Can not set empty password")

        role_list = []
        if roles:
            role_list = [role.strip() for role in roles.split(',')]
        else:
            print "\nSelect role(s) for new user"
            _role_listing()
        role_list = _role_selection(persister=self.persister, choices=role_list)

        try:
            self.persister.begin()
            user_id = User.add_user(username, password, protocol)
            if not role_list:
                print ("You can always assign roles later using the "
                       "'user addrole' command")
            else:
                for role_id in role_list:
                    User.add_user_role(user_id, int(role_id))
        except Exception:
            # Whatever happens, we rollback and re-raise
            self.persister.rollback()
            print "Adding user cancelled."
            raise

        self.persister.commit()
        print "Fabric user added."


class UserDelete(UserCommand):

    """Delete a Fabric user.

    * protocol: Protocol of the user (for example 'xmlrpc')
    * force: Do not ask for confirmation

    """

    command_name = 'delete'
    description = 'Delete a Fabric user'

    def execute(self, username, protocol=None, force=False):
        """Delete a Fabric user
        """
        username, password, protocol = self._ask_credentials(
            username, ask_password=False)

        # Check if user exists
        if not get_user(username, protocol, persister=self.persister):
            raise _errors.CredentialError(
                "No user {user}/{protocol}".format(
                    user=username, protocol=protocol))

        if not self.options.force:
            result = confirm(
                "Really remove user {user}/{protocol}?".format(
                    user=username, protocol=protocol), default='n')
            if result is not True:
                return

        try:
            self.persister.begin()
            User.delete_user(username=username, protocol=protocol)
        except Exception:
            # We rollback and re-raise
            self.persister.rollback()
            print "Removing user cancelled."
            raise

        self.persister.commit()
        print "Fabric user deleted."


class UserPassword(UserCommand):

    """Change password of a Fabric user.

    * protocol: Protocol of the user (for example 'xmlrpc')

    """

    command_name = 'password'
    description = 'Change password a Fabric user'

    def execute(self, username, protocol=None):
        """Change password a Fabric user
        """
        username, password, protocol = self._ask_credentials(
            username, ask_password=False)

        # Check if user exists
        if not get_user(username, protocol):
            raise _errors.CredentialError(
                "No user {user}/{protocol}".format(
                    user=username, protocol=protocol))

        password = _get_password('New password: ')

        if password:
            try:
                _change_password(username, password, protocol, self.config,
                                 self.persister)
            except _errors.CredentialError as error:
                print "Changing password cancelled."
                _LOGGER.debug(str(error))
                print error
            else:
                print "Password changed."
        else:
            raise _errors.CredentialError("Can not set empty password")


class UserRoles(UserCommand):

    """Change roles for a Fabric user

    * protocol: Protocol of the user (for example 'xmlrpc')
    * roles: Comma separated list of roles, IDs or names (see `role list`)

    """

    command_name = 'roles'
    description = 'Change roles for a Fabric user'

    def execute(self, username, protocol=None, roles=None):
        """Change roles for a Fabric user
        """
        username, password, protocol = self._ask_credentials(
            username, ask_password=False)

        # Check if user exists
        user = get_user(username, protocol)
        if not user:
            raise _errors.CredentialError(
                "No user {user}/{protocol}".format(user=username,
                                                   protocol=protocol))

        exit_text_removed = (
            "Roles for {user}/{protocol} removed."
        ).format(user=username, protocol=protocol)
        exit_text_updated = (
            "Roles for {user}/{protocol} updated."
        ).format(user=username, protocol=protocol)

        confirmed = False
        role_list = []
        if not roles:
            options = {
                "raw": False,
                "fetch": False,
                "params": (user.user_id,),
                "columns": True,
            }
            cur = self.persister.exec_stmt(
                "SELECT role_id FROM user_roles WHERE user_id = %s",
                options)
            current_roles = [row.role_id for row in cur]

            print "\nSelect new role(s) for user, replacing current roles."
            print "Current roles are marke with an X."
            _role_listing(selected=current_roles)
            role_list = _role_selection(persister=self.persister)

            if not role_list:
                confirm_text = (
                    "Remove all roles of user {user}/{protocol}?"
                ).format(user=username, protocol=protocol)
                exit_text = exit_text_removed
                default = 'n'
            else:
                confirm_text = (
                    "Replace roles of user {user}/{protocol}?"
                ).format(user=username, protocol=protocol)
                exit_text = exit_text_updated
                default = 'y'

            confirmed = confirm(confirm_text.format(username=user.username),
                                default=default)
        else:
            # From command line option --roles
            confirmed = True
            if roles.strip() == "0":
                exit_text = exit_text_removed
                role_list = []
            else:
                exit_text = exit_text_updated
                role_list = [role.strip() for role in roles.split(',')]

                role_list = _role_selection(persister=self.persister,
                                            choices=role_list)

        if confirmed:
            try:
                self.persister.begin()
                options = {
                    "raw": False,
                    "fetch": False,
                    "params": (user.user_id,),
                    "columns": True,
                }
                self.persister.exec_stmt(
                    "DELETE FROM user_roles WHERE user_id = %s", options)
                for role_id in role_list:
                    User.add_user_role(user.user_id, int(role_id))
            except Exception:
                # Whatever happens, we rollback and re-raise
                self.persister.rollback()
                print "Changing roles for user cancelled."
                raise
            else:
                self.persister.commit()
                print exit_text
        else:
            print "Changing roles cancelled."


class UserList(UserCommand):
    """List users and their roles
    """

    group_name = 'user'
    command_name = 'list'
    description = 'List roles and their permissions'

    def execute(self):
        """Display list of users
        """
        persister = _persistence.current_persister()

        options = {
            "raw": False,
            "fetch": False,
            "columns": True,
        }

        role_perms = (
            "SELECT u.username, u.protocol, r.name as role_name, "
            "r.description AS role_desc,"
            "p.permission_id, p.description AS perm_desc, p.subsystem, "
            "p.component, p.function "
            "FROM users as u LEFT JOIN user_roles AS ur USING (user_id) "
            "LEFT JOIN roles AS r USING (role_id) "
            "LEFT JOIN role_permissions USING (role_id) "
            "LEFT JOIN permissions AS p USING (permission_id) "
            "ORDER BY u.username, u.protocol"
        )
        cur = persister.exec_stmt(role_perms, options)
        roles = {}
        max_username_len = 0
        max_protocol_len = 0

        user_info = {}
        for row in cur:
            if len(row.username) > max_username_len:
                max_username_len = len(row.username)
            if len(row.protocol) > max_protocol_len:
                max_protocol_len = len(row.protocol)

            user_tuple = (row.username, row.protocol)
            if user_tuple not in user_info:
                user_info[user_tuple] = []
            if row.role_name and row.role_name not in user_info[user_tuple]:
                user_info[user_tuple].append(row.role_name)

        # Minimum sizes
        max_username_len = max(9, max_username_len)
        max_protocol_len = max(12, max_protocol_len)

        role_fmt = ("{{username:{userlen}}} {{protocol:{protlen}}} "
                    "{{roles}}".format(userlen=max_username_len,
                                       protlen=max_protocol_len))

        header = role_fmt.format(username="Username", protocol="Protocol",
                                 roles="Roles")
        print header
        print role_fmt.format(username='-' * max_username_len,
                              protocol='-' * max_protocol_len,
                              roles='-' * 20)

        for user_tuple, roles in user_info.iteritems():
            username, protocol = user_tuple
            if roles:
                role_list = ', '.join(roles)
            else:
                role_list = '(no roles set)'
            print role_fmt.format(username=username,
                                  protocol=protocol,
                                  roles=role_list)


class RoleList(UserCommand):
    """List roles and associated permissions
    """

    group_name = 'role'
    command_name = 'list'
    description = 'List roles and their permissions'

    def execute(self):
        _role_listing(persister=self.persister)


def check_credentials(group, command, config, protocol):
    """Check credentials using configuration

    :raises errors.CredentialError: if login failed, or if user has no
                                    permission
    """
    if group not in ('user', 'role'):
        return

    _configure_connections(config)
    _persistence.init_thread()

    if not protocol:
        protocol = FABRIC_DEFAULT_PROTOCOL

    section = 'protocol.' + protocol

    username = config.get(section, 'user')
    password = config.get(section, 'password')
    realm = config.get(section, 'realm', vars=FABRIC_PROTOCOL_DEFAULTS)

    user = User.fetch_user(username, protocol=protocol, realm=realm)
    password_hash = _hash_password(username, password, protocol, config, realm)

    if user is None or user.password_hash != password_hash:
        _LOGGER.info("Failed login for user %s/%s", username, protocol)
        raise _errors.CredentialError("Login failed")
    elif not user.has_permission('core', group, command):
        _LOGGER.info("Permission denied for user %s/%s", username, protocol)
        raise _errors.CredentialError("No permission")



#!/usr/bin/env python
#
# Copyright (c) 2012, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the mysql utilities client.
"""

from mysql.utilities.common.options import license_callback, UtilitiesParser
from mysql.utilities.common.tools import (check_connector_python,
                                          check_python_version)

# Check Python version compatibility
check_python_version()

import os
import sys

# Check for connector/python
if not check_connector_python():
    sys.exit(1)

try:
    from mysql.utilities import VERSION_FRM, VERSION_STRING, COPYRIGHT_FULL
    from mysql.utilities.exception import UtilError
    from mysql.utilities.command.utilitiesconsole import UtilitiesConsole
    from mysql.utilities.common.options import add_verbosity, check_verbosity
except:
    print("ERROR: MySQL Utilities are either not installed or are not "
          "accessible from this terminal.")
    sys.exit(2)

# Constants
NAME = "MySQL Utilities Client - mysqluc "
DESCRIPTION = "mysqluc - Command line client for running MySQL Utilities"
USAGE = "%prog "
WELCOME_MESSAGE = """
Welcome to the MySQL Utilities Client (mysqluc) version {0}\n{1}
Type 'help' for a list of commands or press TAB twice for list of utilities.
"""
GOODBYE_MESSAGE = "\nThanks for using the MySQL Utilities Client!\n"
PRINT_WIDTH = 75
UTIL_PATH = "/scripts"

if __name__ == '__main__':
    def build_variable_dictionary_list(args):
        """Build a variable dictionary from the arguments

        Returns list - list of variables
        """
        variables = []
        arguments = list(args)
        for arg in arguments[:]:
            if '=' in arg:
                try:
                    name, value = arg.split('=')
                    if not value:
                        raise ValueError
                except ValueError:
                    parser.error("Invalid argument assignment: {0}. Please "
                                 "check your command.".format(arg))
                variables.append({'name': name, 'value': value})
                arguments.remove(arg)

        if len(arguments) > 0:
            parser.error("Unbalanced arguments. Please check your command.")
        for i in range(0, len(arguments), 2):
            variables.append({'name': arguments[i], 'value': arguments[i + 1]})
        return variables

    # Setup the command parser
    program = os.path.basename(sys.argv[0]).replace(".py", "")
    parser = UtilitiesParser(
        version=VERSION_FRM.format(program=program),
        description=DESCRIPTION,
        usage=USAGE,
        add_help_option=False,
        prog=program
    )

    # Default option to provide help information
    parser.add_option("--help", action="help",
                      help="display this help message and exit")

    # Add --License option
    parser.add_option("--license", action='callback',
                      callback=license_callback,
                      help="display program's license and exit")

    # Add display width option
    parser.add_option("--width", action="store", dest="width",
                      type="int", help="display width",
                      default=PRINT_WIDTH)

    # Add utility directory option
    parser.add_option("--utildir", action="store", dest="utildir",
                      type="string", help="location of utilities",
                      default=UTIL_PATH)

    # Add execute mode
    parser.add_option("-e", "--execute", action="store", dest="commands",
                      type="string",
                      help="execute commands and exit. Multiple commands are "
                           "separated with semi-colons. Note: some platforms "
                           "may require double quotes around command list.",
                      default=None)

    # Add utility extra_utilities option
    parser.add_option("--add-utility", action="append", dest="add_util",
                      help="append an utility in the format mysql<utility_"
                           "name>. The mysql<utility_name>.py must be located "
                           "inside the folder given by the utildir option",
                      default=[])

    # Add utility extra_utilities option
    parser.add_option("--hide-utils", action="store_true",
                      dest="hide_util",
                      help="when this option is given, the default utilities "
                           "will not be available, must be used only along "
                           "of --add-utility option",
                      default=False)

    # Add verbosity mode
    add_verbosity(parser, True)

    # Now we process the rest of the arguments.
    opt, args = parser.parse_args()

    # Warn if quiet and verbosity are both specified
    check_verbosity(opt)

    if opt.verbosity is None:
        verbosity = 0
    else:
        verbosity = opt.verbosity

    quiet = opt.quiet
    if opt.quiet is None:
        quiet = False

    if opt.hide_util and not opt.add_util:
        # TODO: move to common/messages.py
        parser.error("You can only use --hide_utils option along the "
                     "--add-util option.")

    extra_utils_dict = {}
    for utility in opt.add_util:
        extra_utils_dict[utility] = ()

    options = {
        'verbosity': verbosity,
        'quiet': quiet,
        'width': opt.width,
        'utildir': opt.utildir,
        'variables': build_variable_dictionary_list(args),
        'prompt': 'mysqluc> ',
        'welcome': WELCOME_MESSAGE.format(VERSION_STRING, COPYRIGHT_FULL),
        'goodbye': GOODBYE_MESSAGE,
        'commands': opt.commands,
        'custom': True,  # We are using custom commands
        'hide_util': opt.hide_util,
        'add_util': extra_utils_dict
    }

    try:
        print("Launching console ...")
        util_con = UtilitiesConsole(options)
        util_con.run_console()
    except KeyboardInterrupt:
        print(options['goodbye'])
    except UtilError:
        _, e, _ = sys.exc_info()
        print("ERROR: %s" % e.errmsg)

    sys.exit()



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Define interfaces to manage servers, specifically MySQL Servers.

A server is uniquely identified through a *UUID* (Universally Unique
Identifier) and has an *Address* (i.e., Hostname:Port) which is
used to connect to it through the Python Database API. If a server
process such as MySQL already provides a uuid, the server's class
used to create a MySQL object must ensure that they match otherwise
the different uuids may cause problems in other modules.

Any sort of provisioning must not be performed when the server object
is instantiated. The provisioning steps must be done in other modules.

Servers are organized into groups which have unique names. This aims
at defining administrative domains and easing management activities.
"""
import threading
import uuid as _uuid
import logging
import functools
import re

from datetime import (
    datetime,
)

from mysql.fabric import (
    errors as _errors,
    persistence as _persistence,
    server_utils as _server_utils,
    utils as _utils,
    failure_detector as _detector,
    error_log as _error_log,
    config as _config,
)

from mysql.fabric.handler import (
    MySQLHandler,
)

_LOGGER = logging.getLogger(__name__)

def server_logging(function):
    """This logs information on functions being called within server
    instances.
    """
    @functools.wraps(function)
    def wrapper_check(*args, **kwrds):
        """Inner function that logs information on wrapped function.
        """
        _LOGGER.debug(
            "Start executing function: %s(%s, %s).", function.__name__,
            args, kwrds
        )
        try:
            ret = function(*args, **kwrds)
        except Exception:
            _LOGGER.debug("Error executing function: %s.", function.__name__)
            raise
        else:
            _LOGGER.debug("Finish executing function: %s.", function.__name__)
        return ret
    return wrapper_check


class Group(_persistence.Persistable):
    """Provide interfaces to organize servers into groups.

    :param group_id: The id that uniquely identifies the group.
    :param description: The group's description.
    :param master: The master's uuid in the group.
    :rtype master: UUID
    :param status: Group's status.
    :rtype status: ACTIVE or INACTIVE.
    """
    CREATE_GROUP = ("CREATE TABLE groups"
                    "(group_id VARCHAR(64) NOT NULL, "
                    "description VARCHAR(256), "
                    "master_uuid VARCHAR(40), "
                    "master_defined TIMESTAMP(6) NULL, "
                    "status BIT(1) NOT NULL, "
                    "CONSTRAINT pk_group_id PRIMARY KEY (group_id))")

    #Create the table that stores the group replication relationship.
    CREATE_GROUP_REPLICATION = (
                            "CREATE TABLE group_replication"
                            "(master_group_id VARCHAR(64) NOT NULL, "
                            "slave_group_id VARCHAR(64) NOT NULL, "
                            "CONSTRAINT pk_master_slave_group_id "
                            "PRIMARY KEY(master_group_id, slave_group_id), "
                            "CONSTRAINT FOREIGN KEY(master_group_id) "
                            "REFERENCES groups(group_id), "
                            "CONSTRAINT FOREIGN KEY(slave_group_id) "
                            "REFERENCES groups(group_id), "
                            "INDEX idx_slave_group_id(slave_group_id))")

    #SQL statement for inserting a new group into the table
    INSERT_GROUP = ("INSERT INTO groups(group_id, description, status) "
                    "VALUES(%s, %s, %s)")

    #SQL statement for selecting all groups
    QUERY_GROUPS = ("SELECT group_id FROM groups")

    #SQL statement for selecting all groups
    QUERY_GROUPS_BY_STATUS = ("SELECT group_id FROM groups WHERE status = %s")

    #Query the group that is the master of this group.
    QUERY_GROUP_REPLICATION_MASTER = ("SELECT master_group_id FROM "
                                "group_replication WHERE slave_group_id = %s")

    #Query the groups that are the slaves of this group.
    QUERY_GROUP_REPLICATION_SLAVES = ("SELECT slave_group_id FROM "
                                "group_replication WHERE master_group_id = %s")

    #SQL statement for updating the group table identified by the group id.
    UPDATE_GROUP = ("UPDATE groups SET description = %s WHERE group_id = %s")

    #SQL statement used for deleting the group identified by the group id.
    REMOVE_GROUP = ("DELETE FROM groups WHERE group_id = %s")

    #Remove the mapping between a master and a slave group.
    DELETE_MASTER_SLAVE_GROUP_MAPPING = ("DELETE FROM group_replication "
                            "WHERE slave_group_id = %s")

    #Delete all the master to slave mappings for a given master group. A
    #given master group can have multiple slaves.
    DELETE_SLAVE_GROUPS = ("DELETE FROM group_replication "
                            "WHERE master_group_id = %s")

    #Add a Master - Slave Group mapping.
    INSERT_MASTER_SLAVE_GROUP_MAPPING = \
            ("INSERT INTO group_replication"
             "(master_group_id, slave_group_id)"
             " VALUES(%s, %s)")

    #SQL Statement to retrieve a specific group from the state_store.
    QUERY_GROUP = ("SELECT group_id, description, master_uuid, "
                   "master_defined, status FROM groups WHERE group_id = %s")

    #SQL Statement to update the group's master.
    UPDATE_MASTER = ("UPDATE groups SET master_uuid = %s, master_defined = %s "
                     "WHERE group_id = %s")

    #SQL Statement to update the group's status.
    UPDATE_STATUS = ("UPDATE groups SET status = %s WHERE group_id = %s")

    #Create the referential integrity constraint with the servers table
    ADD_FOREIGN_KEY_CONSTRAINT_MASTER_UUID = (
        "ALTER TABLE groups ADD CONSTRAINT fk_master_uid_servers "
        "FOREIGN KEY(master_uuid) REFERENCES servers(server_uuid)"
    )

    #Group's statuses
    INACTIVE, ACTIVE = range(0, 2)

    #List with Group's statuses
    GROUP_STATUS = [INACTIVE, ACTIVE]

    #Failover interval
    _FAILOVER_INTERVAL = _DEFAULT_FAILOVER_INTERVAL = 3600

    def __init__(self, group_id, description=None, master=None,
                 master_defined=None, status=INACTIVE):
        """Constructor for the Group.
        """
        assert(isinstance(group_id, basestring))
        assert(description is None or isinstance(description, basestring))
        assert(master is None or isinstance(master, _uuid.UUID))
        assert(master_defined is None or isinstance(master_defined, datetime))
        assert(status in Group.GROUP_STATUS)
        super(Group, self).__init__()
        self.__group_id = group_id
        self.__description = description
        self.__master = master
        self.__master_defined = master_defined
        self.__status = status

    def __eq__(self,  other):
        """Two groups are equal if they have the same id.
        """
        return isinstance(other, Group) and \
               self.__group_id == other.group_id

    def __hash__(self):
        """A group is hashable through its uuid.
        """
        return hash(self.__group_id)

    @property
    def group_id(self):
        """Return the group's id.
        """
        return self.__group_id

    @property
    def slave_group_ids(self):
        """Property that gives access to the list of Groups that are
        slaves to this group.
        """
        return self.fetch_slave_group_ids()

    def fetch_slave_group_ids(self, persister=None):
        """Return the list of Groups that are slaves to this group.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        ret = []
        rows = persister.exec_stmt(Group.QUERY_GROUP_REPLICATION_SLAVES,
            {"fetch" : True, "params" : (self.__group_id,)})
        if not rows:
            return ret

        for row in rows:
            ret.append(row[0])
        return ret

    @property
    def master_group_id(self):
        """Property that returns the ID of the master group from which this
        group replicates.
        """
        return self.fetch_master_group_id()

    def fetch_master_group_id(self, persister=None):
        """Return the ID of the master group from which this group replicates.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        row = persister.exec_stmt(Group.QUERY_GROUP_REPLICATION_MASTER,
            {"fetch" : True, "params" : (self.__group_id,)})
        if not row:
            return None
        return row[0][0]


    def add_slave_group_id(self,  slave_group_id, persister=None):
        """Insert a slave group ID into the slave group ID list. Register a
        slave to this group.

        :param slave_group_id: the group ID of the slave group that needs to
                                              be added.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Group.INSERT_MASTER_SLAVE_GROUP_MAPPING,
                            {"params": (self.__group_id, slave_group_id)})

    def remove_slave_group_id(self,  slave_group_id, persister=None):
        """Remove a slave group ID from the slave group ID list. Unregister a
        slave group.

        :param slave_group_id: the group ID of the slave group that needs to
                                              be removed.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Group.DELETE_MASTER_SLAVE_GROUP_MAPPING,
                            {"params": (slave_group_id, )})

    def remove_slave_group_ids(self, persister=None):
        """Remove slave group ids for a particular group. Unregisters
        all the slave of this group.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Group.DELETE_SLAVE_GROUPS,
                            {"params": (self.__group_id, )})

    def add_master_group_id(self,  master_group_id, persister=None):
        """Set the master group ID. Register a group as a master. This Group
        basically is a slave to the registered group.

        :param master_group_id: The group ID of the master that needs to be
                                                 added.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Group.INSERT_MASTER_SLAVE_GROUP_MAPPING,
                            {"params": (master_group_id, self.__group_id)})

    def remove_master_group_id(self, persister=None):
        """Remove the master group ID. Unregister a master group.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Group.DELETE_MASTER_SLAVE_GROUP_MAPPING,
                            {"params": (self.__group_id, )})

    def add_server(self, server):
        """Add a server into this group.

        :param server: The Server object that needs to be added to this
                       Group.
        """
        assert(isinstance(server, MySQLServer))
        assert(server.group_id == None)
        server.group_id = self.__group_id

    def remove_server(self, server):
        """Remove a server from this group.

        :param server: The Server object that needs to be removed from this
                       Group.
        """
        assert(isinstance(server, MySQLServer))
        assert(server.group_id == self.__group_id)
        server.group_id = None

    @property
    def description(self):
        """Return the description for the group.
        """
        return self.__description

    @description.setter
    def description(self, description=None, persister=None):
        """Set the description for this group. Update the description for the
        Group in the state store.

        :param persister: The DB server that can be used to access the
                          state store.
        :param description: The new description for the group that needs to be
                            updated.
        """
        persister.exec_stmt(Group.UPDATE_GROUP,
            {"params":(description, self.__group_id)})
        self.__description = description

    @property
    def master(self):
        """Return the master for the group.
        """
        return self.__master

    @property
    def master_defined(self):
        """Return the last time the master has changed.
        """
        return self.__master_defined

    @master.setter
    def master(self, master, persister=None):
        """Set the master for this group.

        :param persister: The DB server that can be used to access the
                          state store.
        :param master: The master for the group that needs to be updated.
        """
        assert(master is None or isinstance(master, _uuid.UUID))
        if master is None:
            param_master = None
        else:
            param_master = str(master)

        _LOGGER.info("Master has changed from %s to %s.", self.__master, master,
            extra={
                "subject": self.__group_id,
                "category": MySQLHandler.GROUP,
                "type" : MySQLHandler.PROMOTE if master else \
                         MySQLHandler.DEMOTE
            }
        )
        persister.exec_stmt(Group.UPDATE_MASTER,
            {"params":(param_master, _utils.get_time(), self.__group_id)})
        self.__master = master


    def servers(self):
        """Return a list with the servers in this group.
        """
        return MySQLServer.servers(self.__group_id)

    @property
    def status(self):
        """Return the group's status.
        """
        return self.__status

    @status.setter
    def status(self, status, persister=None):
        """Set the group's status.

        :param status: The new group's status.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        assert(status in Group.GROUP_STATUS)
        # Check the maximum number of threads.
        _utils.check_number_threads(1)
        persister.exec_stmt(Group.UPDATE_STATUS,
            {"params":(status, self.__group_id)})
        self.__status = status

    def can_set_server_faulty(self, server, now):
        """Check whether is ipossible to set a new master.

        If `now - master_defined` > Group._FAILOVER_INTERVAL, it is safe
        to set a new master without making the system unstable.
        """
        if self.__master_defined is None:
            return True

        diff = now - self.__master_defined
        interval = _utils.get_time_delta(Group._FAILOVER_INTERVAL)

        if (self.__master == server.uuid and diff >= interval) or \
            self.__master != server.uuid:
            return True

        return False

    @staticmethod
    def groups_by_status(status, persister=None):
        """Return the group_ids of all the available groups.

        :param persister: Persister to persist the object to.
        """
        assert(status in Group.GROUP_STATUS)
        return persister.exec_stmt(Group.QUERY_GROUPS_BY_STATUS,
            {"params":(status, )}
            )

    @staticmethod
    def groups(persister=None):
        """Return the group_ids of all the available groups.

        :param persister: Persister to persist the object to.
        """
        return persister.exec_stmt(Group.QUERY_GROUPS)

    @staticmethod
    def remove(group, persister=None):
        """Remove the group object from the state store.

        :param group: A reference to a group.
        :param persister: Persister to persist the object to.
        """
        persister.exec_stmt(
            Group.REMOVE_GROUP, {"params" : (group.group_id, )}
            )

    @staticmethod
    def fetch(group_id, persister=None):
        """Return the group object, by loading the attributes for the group_id
        from the state store.

        :param group_id: The group_id for the Group object that needs to be
                         retrieved.
        :param persister: Persister to persist the object to.
        :return: The Group object corresponding to the group_id
                 None if the Group object does not exist.
        """
        group = None
        cur = persister.exec_stmt(
            Group.QUERY_GROUP, {"fetch" : False, "raw" : False,
            "params" : (group_id,)}
            )
        row = cur.fetchone()
        if row:
            group_id, description, master, master_defined, status = row
            if master:
                master = _uuid.UUID(master)
            group = Group(
                group_id=group_id, description=description, master=master,
                master_defined=master_defined, status=status
                )
        return group

    @staticmethod
    def add(group, persister=None):
        """Write a group object into the state store.

        :param group: A reference to a group.
        :param persister: Persister to persist the object to.
        """
        persister.exec_stmt(Group.INSERT_GROUP,
            {"params": (group.group_id, group.description, group.status)}
            )

    @staticmethod
    def create(persister=None):
        """Create the objects(tables) that will store the Group information in
        the state store.

        :param persister: The DB server that can be used to access the
                          state store.
        :raises: DatabaseError If the table already exists.
        """
        persister.exec_stmt(Group.CREATE_GROUP)
        persister.exec_stmt(Group.CREATE_GROUP_REPLICATION)

    @staticmethod
    def add_constraints(persister=None):
        """Add the constraints to the groups table.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(
                Group.ADD_FOREIGN_KEY_CONSTRAINT_MASTER_UUID
        )
        return True


class ConnectionPool(_utils.Singleton):
    """Manages MySQL Servers' connections.

    The pool is internally implemented as a dictionary that maps a server's
    uuid to a sequence of connections.
    """
    def __init__(self):
        """Creates a ConnectionPool object.
        """
        super(ConnectionPool, self).__init__()
        self.__pool = {}
        self.__lock = threading.RLock()

    def get_connection(self, uuid, user):
        """Get a connection.

        The method gets a connection from a pool if there is any.
        """
        assert(isinstance(uuid, _uuid.UUID))
        with self.__lock:
            try:
                while len(self.__pool[uuid]):
                    cnx = self.__pool[uuid].pop()
                    assert(user != None and cnx.user == user)
                    if _server_utils.is_valid_mysql_connection(cnx):
                        return cnx
            except (KeyError, IndexError):
                pass
        return None

    def release_connection(self, uuid, cnx):
        """Release a connection to the pool.

        It is up to the developer to check if the connection is still
        valid and belongs to this server before returning it to the pool.
        """
        assert(isinstance(uuid, _uuid.UUID))
        with self.__lock:
            if uuid not in self.__pool:
                self.__pool[uuid] = []
            self.__pool[uuid].append(cnx)

    def get_number_connections(self, uuid):
        """Return the number of connections available in the pool.
        """
        assert(isinstance(uuid, _uuid.UUID))
        with self.__lock:
            try:
                return len(self.__pool[uuid])
            except KeyError:
                pass
        return 0

    def purge_connections(self, uuid):
        """Close and remove all connections that belongs to a MySQL Server
        which is identified by its uuid.
        """
        assert(isinstance(uuid, _uuid.UUID))
        with self.__lock:
            try:
                for cnx in self.__pool[uuid]:
                    _server_utils.destroy_mysql_connection(cnx)
                del self.__pool[uuid]
            except KeyError:
                pass


class MySQLServer(_persistence.Persistable):
    """Proxy class that provides an interface to access a MySQL Server
    Instance.

    To create a MySQLServer object, one needs to provide at least two
    parameters: uuid and address (i.e., host:port). If the uuid is not known
    beforehand, one can find it out as follows::

      address = "localhost:13000"
      uuid = MySQLServer.discover_uuid(address)
      uuid = _uuid.UUID(uuid)

      server = MySQLServer(uuid=uuid, address=address)

    After creating the object, it is necessary to connect it to the MySQL
    Server by explicitly calling connect(). This is required because all the
    necessary information to connect to the server may not have been defined
    at initialization time. For example, one may have created the object
    before reading its state from a persistence layer and setting them.

    So after connecting the object to a server, users may execute statements
    by calling exec_stmt() as follows::

      server.connect()
      ret = server.exec_stmt("SELECT VERSION()")
      print "MySQL Server has version", ret[0][0]

    Changing the value of the properties user or password triggers a call to
    disconnect.

    :param uuid: The uuid of the server.
    :param address:  The address of the server.
    :param user: The user's name used to access the server.
    :param passwd: The password used to access the server.
    :param mode: Server's mode.
    :type mode: OFFLINE, READ_ONLY, READ_WRITE.
    :param status: Server's status.
    :type status: FAULTY, SPARE, SECONDARY, PRIMARY.
    :param weight: Server's weight which determines its likelihood of
                   receiving requests.
    :type weight: Float
    :param group_id: Group's id which the server belongs to.
    """
    #SQL Statement for creating the table used to store details about the
    #server.
    CREATE_SERVER = (
        "CREATE TABLE servers "
        "(server_uuid VARCHAR(40) NOT NULL, "
        "server_address VARCHAR(128) NOT NULL, "
        "mode INTEGER NOT NULL, "
        "status INTEGER NOT NULL, "
        "weight FLOAT NOT NULL, "
        "group_id VARCHAR(64), "
        "CONSTRAINT pk_server_uuid PRIMARY KEY (server_uuid), "
        "INDEX idx_group_id (group_id))"
    )

    #Create the referential integrity constraint with the groups table
    ADD_FOREIGN_KEY_CONSTRAINT_GROUP_ID = (
        "ALTER TABLE servers ADD CONSTRAINT fk_group_id_servers "
        "FOREIGN KEY(group_id) REFERENCES groups(group_id)"
    )

    #SQL statement for inserting a new server into the table
    INSERT_SERVER = ("INSERT INTO servers(server_uuid, server_address, mode, "
                     "status, weight, group_id) values(%s, %s, %s, %s, %s, %s)")

    #SQL statement for updating the server table identified by the server id.
    UPDATE_SERVER_MODE = (
        "UPDATE servers SET mode = %s WHERE server_uuid = %s"
        )

    #SQL statement for updating the server table identified by the server id.
    UPDATE_SERVER_STATUS = (
        "UPDATE servers SET status = %s WHERE server_uuid = %s"
        )

    #SQL statement for updating the server table identified by the server id.
    UPDATE_SERVER_WEIGHT = (
        "UPDATE servers SET weight = %s WHERE server_uuid = %s"
        )

    #SQL statement for updating the server table identified by the server id.
    UPDATE_SERVER_GROUP_ID = (
        "UPDATE servers SET group_id = %s WHERE server_uuid = %s"
        )

    #SQL statement used for deleting the server identified by the server id.
    REMOVE_SERVER = ("DELETE FROM servers WHERE server_uuid = %s")

    #SQL Statement to retrieve the server from the state_store.
    QUERY_SERVER = (
        "SELECT server_uuid, server_address, mode, status, weight, group_id "
        "FROM servers WHERE server_uuid = %s"
        )

    #SQL Statement to retrieve a set of servers in a group from the state_store.
    QUERY_SERVER_GROUP_ID = (
        "SELECT server_uuid, server_address, mode, status, weight "
        "FROM servers WHERE group_id = %s"
        )

    #SQL Statement to retrieve the servers belonging to a group.
    DUMP_SERVERS = (
        "SELECT server_uuid, group_id, server_address, mode, status, weight "
        "FROM servers WHERE group_id LIKE %s AND group_id IS NOT NULL AND "
        "status != %s ORDER BY group_id, server_address, server_uuid"
        )

    #Default weight for the server
    DEFAULT_WEIGHT = 1.0

    # Define a session context for a variable.
    SESSION_CONTEXT = "SESSION"

    # Define a global context for a variable.
    GLOBAL_CONTEXT = "GLOBAL"

    # Set of contexts.
    CONTEXTS = [SESSION_CONTEXT, GLOBAL_CONTEXT]

    # Define the offline mode.
    OFFLINE = "OFFLINE"

    # Define the read-only mode.
    READ_ONLY = "READ_ONLY"

    # Define the write-only mode.
    WRITE_ONLY = "WRITE_ONLY"

    # Define the read-write mode.
    READ_WRITE = "READ_WRITE"

    # Define default mode
    DEFAULT_MODE = READ_ONLY

    # Set of possible modes.
    SERVER_MODE = [OFFLINE, READ_ONLY, WRITE_ONLY, READ_WRITE]

    # Define the faulty status.
    FAULTY = "FAULTY"

    # Define the spare status.
    SPARE = "SPARE"

    # Define the secondary status.
    SECONDARY = "SECONDARY"

    # Define the primary status.
    PRIMARY = "PRIMARY"

    # Define default status
    DEFAULT_STATUS = SECONDARY

    # Set of possible statuses.
    SERVER_STATUS = [FAULTY, SPARE, SECONDARY, PRIMARY]

    USER = None

    PASSWD = None

    ALL_PRIVILEGES = [ "ALL PRIVILEGES" ]

    NO_USER_DATABASES = ["performance_schema", "information_schema", "mysql"]

    def __init__(self, uuid, address, user=None, passwd=None, mode=DEFAULT_MODE,
                 status=DEFAULT_STATUS, weight=DEFAULT_WEIGHT, group_id=None):
        """Constructor for MySQLServer.
        """
        super(MySQLServer, self).__init__()
        assert(isinstance(uuid, _uuid.UUID))
        assert(mode in MySQLServer.SERVER_MODE)
        assert(status in MySQLServer.SERVER_STATUS)
        self.__cnx = None
        self.__uuid = uuid
        self.__address = address
        self.__group_id = group_id
        self.__pool = ConnectionPool()
        self.__user = user
        self.__passwd = passwd
        self.__mode = mode
        self.__status = status
        self.__weight = weight
        self.__read_only = None
        self.__server_id = None
        self.__version = None
        self.__gtid_enabled = None
        self.__binlog_enabled = None

    @staticmethod
    @server_logging
    def discover_uuid(address, user=None, passwd=None,
                      connection_timeout=None):
        """Retrieve the uuid from a server.

        :param address: Server's address.
        :param user: The user's name used to access the server.
        :param passwd: The password used to access the server.
        :param connection_timeout: Time in seconds after which an error is
                                   reported if the UUID is not retrieved.
        :return: UUID.
        """

        host, port = _server_utils.split_host_port(
            address,  _server_utils.MYSQL_DEFAULT_PORT
        )
        port = int(port)

        user = user or MySQLServer.USER
        passwd = passwd or MySQLServer.PASSWD
        cnx = _server_utils.create_mysql_connection(
            user=user, passwd=passwd, host=host, port=port, autocommit=True,
            use_unicode=False, connection_timeout=connection_timeout
        )

        try:
            row = _server_utils.exec_mysql_stmt(cnx,
                "SELECT @@GLOBAL.SERVER_UUID")
            server_uuid = row[0][0]
        finally:
            _server_utils.destroy_mysql_connection(cnx)

        return server_uuid

    def _do_connection(self):
        """Get a connection from the pool provided there is one or create
        a fresh connection.
        """
        cnx = self.__pool.get_connection(self.__uuid, self.user)
        if cnx:
            return cnx

        return self._do_create_connection()

    def _do_create_connection(self, connection_timeout=None):
        """Create a new connection.
        """
        host, port = _server_utils.split_host_port(self.address,
            _server_utils.MYSQL_DEFAULT_PORT)

        return _server_utils.create_mysql_connection(
            autocommit=True, use_unicode=False, host=host, port=port,
            user=self.user, passwd=self.passwd,
            connection_timeout=connection_timeout
        )

    def connect(self):
        """Connect to a MySQL Server instance.
        """
        self.disconnect()

        # Set up an internal connection.
        self.__cnx = self._do_connection()
        _LOGGER.debug("Using connection (%s).", self.__cnx)

        # Get server's uuid
        ret_uuid = self.get_variable("SERVER_UUID")
        ret_uuid = _uuid.UUID(ret_uuid)
        if ret_uuid != self.uuid:
            self.disconnect()
            raise _errors.UuidError(
                "UUIDs do not match (stored (%s), read (%s))." %
                (self.uuid, ret_uuid)
            )

        # Get server's id.
        self.__server_id = int(self.get_variable("SERVER_ID"))

        # Get server's version.
        self.__version = self.get_variable("VERSION")

        # Get information on gtid support.
        if not self.check_version_compat((5, 6, 5)):
            self.__gtid_enabled = False
        else:
            ret_gtid = self.get_variable("GTID_MODE")
            self.__gtid_enabled = ret_gtid in ("ON", "1")

        ret_binlog = self.get_variable("LOG_BIN")
        self.__binlog_enabled = not ret_binlog in ("OFF", "0")

        self._check_read_only()

        _LOGGER.debug("Connected to server with uuid (%s), server_id (%d), "
                      "version (%s), gtid (%s), binlog (%s), read_only (%s).",
                      self.uuid, self.__server_id, self.__version,
                      self.__gtid_enabled, self.__binlog_enabled,
                      self.__read_only)

    def disconnect(self):
        """Disconnect from the server.
        """
        if self.__cnx is not None:
            _LOGGER.debug("Disconnecting from server with uuid (%s), "
                          "server_id (%s), version (%s), gtid (%s), "
                          "binlog (%s), read_only (%s).", self.uuid,
                          self.__server_id, self.__version,
                          self.__gtid_enabled, self.__binlog_enabled,
                          self.__read_only)
            self.__pool.release_connection(self.__uuid, self.__cnx)
            self.__cnx = None
            self.__read_only = None
            self.__server_id = None
            self.__version = None
            self.__gtid_enabled = None
            self.__binlog_enabled = None

    def user_databases(self):
        """Return user databases.
        """
        exc_dbs = MySQLServer.NO_USER_DATABASES
        dbs = self.exec_stmt("SHOW DATABASES")
        dbs = [db[0] for db in dbs if db[0] not in exc_dbs]
        return dbs

    def has_privileges(self, required_privileges, level=None):
        """Check whether the current user has the required privileges.

        :param required_privileges: List or tuple of required privileges which
                                    a user who is connected to the current
                                    server must have.
        :param level: Level of the set of privileges.
        """
        assert(isinstance(required_privileges, list) or \
               isinstance(required_privileges, tuple))

        required_privileges = set(required_privileges)
        required_level = level or "*.*"
        all_privileges = "ALL PRIVILEGES"
        all_level = "*.*"
        ret = self.exec_stmt("SHOW GRANTS")
        check = re.compile("GRANT (?P<privileges>.*?) ON (?P<level>.*?) TO")
        for row in ret:
            res = check.match(row[0])
            if res:
                privileges = [ privilege.strip() \
                      for privilege in res.group("privileges").split(",")
                ]
                level = res.group("level").replace('`', "")
                if (all_privileges in privileges and all_level == level) or \
                    (required_privileges.issubset(set(privileges)) and \
                    required_level == level):
                    return True
        return False

    def has_required_privileges(self):
        """Check whether the current user has a minimum set of privileges:
        MySQLServer.ALL_PRIVILEGES on *.*.
        """
        return self.has_privileges(MySQLServer.ALL_PRIVILEGES)

    def is_connected(self):
        """Determine whether the proxy object (i.e. the server) is connected
        to actual server.
        """
        return self.__cnx is not None

    def is_alive(self, connection_timeout=None):
        """Determine whether the server is dead or alive by trying to create
        a new connection.

        :param connection_timeout: Timeout waiting for getting a new
                                   connection.
        """
        res = False
        try:
            cnx = self._do_create_connection(connection_timeout)
            res = True
            _server_utils.destroy_mysql_connection(cnx)
        except _errors.DatabaseError:
            pass

        return res

    def _check_read_only(self):
        """Check if the database was set to read-only mode.
        """
        ret_read_only = self.get_variable("READ_ONLY")
        self.__read_only = not ret_read_only in ("OFF", "0")

    @property
    def uuid(self):
        """Return the server's uuid.
        """
        return self.__uuid

    @property
    def address(self):
        """Return the server's address.
        """
        return self.__address

    @property
    def read_only(self):
        """Check read only mode on/off.

        :return True If read_only is set
                False If read_only is not set.
        """
        return self.__read_only

    @read_only.setter
    def read_only(self, enabled):
        """Turn read only mode on/off. Persist the information in the state
        store.

        :param enabled The read_only flag value.
        """
        self.set_variable("READ_ONLY", "ON" if enabled else "OFF")
        self._check_read_only()

    @property
    def server_id(self):
        """Return the server id.
        """
        return self.__server_id

    @property
    def version(self):
        """Return version number of the server.
        """
        return self.__version

    @property
    def gtid_enabled(self):
        """Return if GTID is enabled.
        """
        return self.__gtid_enabled

    @property
    def binlog_enabled(self):
        """Check binary logging status.
        """
        return self.__binlog_enabled

    @property
    def user(self):
        """Return user's name who is used to connect to a server.
        """
        return self.__user or MySQLServer.USER

    @user.setter
    def user(self, user):
        """Set user's name who is used to connect to a server.

        :param user: User's name.
        """
        if self.__user != user:
            self.disconnect()
            self.__user = user

    @property
    def passwd(self):
        """Return user's password who is used to connect to a server. Load
        the server information from the state store and return the password.
        """
        return self.__passwd or MySQLServer.PASSWD

    @passwd.setter
    def passwd(self, passwd):
        """Set user's passord who is used to connect to a server.

        :param passwd: User's password.
        """
        if self.__passwd != passwd:
            self.disconnect()
            self.__passwd = passwd

    @property
    def mode(self):
        """Return the server's mode.
        """
        return self.__mode

    @mode.setter
    def mode(self, mode, persister=None):
        """Set the server's mode.

        :param mode: The new server's mode.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        assert(mode in MySQLServer.SERVER_MODE)
        idx = MySQLServer.get_mode_idx(mode)
        persister.exec_stmt(MySQLServer.UPDATE_SERVER_MODE,
                            {"params":(idx, str(self.uuid))})
        self.__mode = mode

    @staticmethod
    def get_mode_idx(mode):
        """Return the index associated to a mode.
        """
        return MySQLServer.SERVER_MODE.index(mode)

    @staticmethod
    def get_mode(idx):
        """Return the mode associated to an index.
        """
        return MySQLServer.SERVER_MODE[idx]

    @property
    def status(self):
        """Return the server's status.
        """
        return self.__status

    @status.setter
    def status(self, status, persister=None):
        """Set the server's status.

        :param status: The new server's status.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        assert(status in MySQLServer.SERVER_STATUS)
        idx = MySQLServer.get_status_idx(status)
        persister.exec_stmt(MySQLServer.UPDATE_SERVER_STATUS,
                            {"params":(idx, str(self.uuid))})
        self.__status = status

    @staticmethod
    def get_status_idx(status):
        """Return the index associated to a status.
        """
        return MySQLServer.SERVER_STATUS.index(status)

    @staticmethod
    def get_status(idx):
        """Return the status associated to an index.
        """
        return MySQLServer.SERVER_STATUS[idx]

    @property
    def weight(self):
        """Return the server's weight.
        """
        return self.__weight

    @weight.setter
    def weight(self, weight, persister=None):
        """Set the server's weight.

        :param weight: The new server's weight.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        assert(weight > 0.0)
        persister.exec_stmt(MySQLServer.UPDATE_SERVER_WEIGHT,
                            {"params":(weight, str(self.uuid))})
        self.__weight = weight

    @property
    def group_id(self):
        """Return the server's group_id.
        """
        return self.__group_id

    @group_id.setter
    def group_id(self, group_id, persister=None):
        """Set the server's group_id.

        :param group_id: The new server's group_id.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(MySQLServer.UPDATE_SERVER_GROUP_ID,
                            {"params":(group_id, str(self.uuid))})
        self.__group_id = group_id

    @staticmethod
    def servers(group_id, persister=None):
        """Return a list of servers identified by group_id.

        :param group_id: Group's id.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        ret = []
        rows = persister.exec_stmt(MySQLServer.QUERY_SERVER_GROUP_ID,
            {"raw" : False, "params" : (group_id, )}
        )
        for row in rows:
            uuid, address, idx_mode, idx_status, weight = row
            mode = MySQLServer.get_mode(idx_mode)
            status = MySQLServer.get_status(idx_status)
            server = MySQLServer(
                _uuid.UUID(uuid), address=address, mode=mode, status=status,
                weight=weight, group_id=group_id
            )
            ret.append(server)
        return ret

    def check_version_compat(self, expected_version):
        """Check version of the server against requested version.

        This method can be used to check for version compatibility.

        :param expected_version: Target server version.
        :type expected_version: (major, minor, release)
        :return: True if server version is GE (>=) version specified,
                 False if server version is LT (<) version specified.
        :rtype: Bool
        """
        assert(isinstance(expected_version, tuple))
        index = self.__version.find("-")
        version_str = self.__version[0 : index] \
            if self.__version.find("-") >= 0 else self.__version
        version = tuple(int(part) for part in version_str.split("."))
        return version >= expected_version

    def get_gtid_status(self):
        """Get the GTID information for the server.

        This method attempts to retrieve the GTID lists. If the server
        does not have GTID turned on or does not support GTID, the method
        will throw the exception DatabaseError.

        :return: A named tuple with GTID information.

        In order to access the result set one may do what follows::

          ret = server.get_gtid_status()
          for record in ret:
            print "GTID_EXECUTED", record.GTID_EXECUTED, record[0]
            print "GTID_PURGED", record.GTID_PURGED, record[1]
            print "GTID_OWNED", record_GTID_OWNED, record[2]
        """
        # Check servers for GTID support
        if self.__cnx and not self.__gtid_enabled:
            raise _errors.ProgrammingError("Global Transaction IDs are not "
                                           "supported.")

        query_str = (
            "SELECT @@GLOBAL.GTID_EXECUTED as GTID_EXECUTED, "
            "@@GLOBAL.GTID_PURGED as GTID_PURGED, "
            "@@GLOBAL.GTID_OWNED as GTID_OWNED"
        )

        return self.exec_stmt(query_str, {"columns" : True})

    def has_storage_engine(self, target):
        """Check to see if an engine exists and is supported.

        :param target: Name of engine to find.
        :return: True if engine exists and is active. False if it does
                 not exist or is not supported/not active/disabled.
        """
        if len(target) == 0:
            return True # This says we will use default engine on the server.

        query_str = (
            "SELECT UPPER(engine) as engine, UPPER(support) as support "
            "FROM INFORMATION_SCHEMA.ENGINES"
        )

        if target:
            engines = self.exec_stmt(query_str)
            for engine in engines:
                if engine[0].upper() == target.upper() and \
                   engine[1].upper() in ['YES', 'DEFAULT']:
                    return True
        return False

    def get_binary_logs(self):
        """Return information on binary logs. Look up `SHOW BINARY LOGS` in
        the MySQL Manual for further details.

        :return: A named tuple with information on binary logs.
        """
        return self.exec_stmt("SHOW BINARY LOGS", {"columns" : True})

    def set_session_binlog(self, enabled=True):
        """Enable or disable binary logging for the client.

        :param disable: If 'disable', turn off the binary log
                        otherwise turn binary log on.

        .. note::

           User must have SUPER privilege to execute this.
        """
        self.set_variable("SQL_LOG_BIN", "ON" if enabled else "OFF",
                          MySQLServer.SESSION_CONTEXT)

    def session_binlog_enabled(self):
        """Check if binary logging is enabled for the client.
        """
        ret = self.get_variable("SQL_LOG_BIN",
                                MySQLServer.SESSION_CONTEXT)
        return ret in ["ON", '1']

    def foreign_key_checks_enabled(self):
        """Check foreign key status for the client.
        """
        ret = self.get_variable("FOREIGN_KEY_CHECKS",
                                MySQLServer.SESSION_CONTEXT)
        return ret in ["ON", '1']

    def set_foreign_key_checks(self, enabled=True):
        """Enable or disable foreign key checks for the client.

        :param disable: If True, turn off foreign key checks otherwise turn
                        foreign key checks on.
        """
        self.set_variable("FOREIGN_KEY_CHECKS", "ON" if enabled else "OFF",
                          MySQLServer.SESSION_CONTEXT)

    def get_variable(self, variable, context=None):
        """Execute the SELECT command for the client and return a result set.
        """
        if not context:
            context = MySQLServer.GLOBAL_CONTEXT
        assert(context in MySQLServer.CONTEXTS)
        ret = self.exec_stmt("SELECT @@%s.%s as %s" %
                             (context, variable, variable))
        return ret[0][0]

    def set_variable(self, variable, value, context=None):
        """Execute the SET command for the client and return a result set.
        """
        if not context:
            context = MySQLServer.GLOBAL_CONTEXT
        assert(context in MySQLServer.CONTEXTS)
        return self.exec_stmt("SET @@%s.%s = %s" %
                              (context, variable, value))

    def exec_stmt(self, stmt_str, options=None):
        """Execute statements against the server.
        See :meth:`~mysql.fabric.server_utils.exec_stmt`.
        """
        return _server_utils.exec_mysql_stmt(self.__cnx, stmt_str, options)

    def __del__(self):
        """Destructor for MySQLServer.
        """
        try:
            self.disconnect()
        except AttributeError:
            pass

    @staticmethod
    def remove(server, persister=None):
        """Remove a server from the state store.

        :param server: A reference to a server.
        :param persister: Persister to persist the object to.
        """
        ConnectionPool().purge_connections(server.uuid)
        _error_log.ErrorLog.remove(server)
        persister.exec_stmt(
            MySQLServer.REMOVE_SERVER, {"params": (str(server.uuid), )}
            )

    @staticmethod
    def fetch(uuid, persister=None):
        """Return a server object corresponding to the uuid.

        :param uuid: The server id of the server object that needs to be
                     returned.
        :param persister: Persister to persist the object to.
        :return: The server object that corresponds to the server id
                 None if the server id does not exist.
        """
        cur = persister.exec_stmt(MySQLServer.QUERY_SERVER,
            {"fetch" : False, "raw" : False, "params":(str(uuid),)}
        )
        row = cur.fetchone()
        if row:
            uuid, address, idx_mode, idx_status, weight, group_id = row
            mode = MySQLServer.get_mode(idx_mode)
            status = MySQLServer.get_status(idx_status)
            return MySQLServer(
                _uuid.UUID(uuid), address=address, mode=mode, status=status,
                weight=weight, group_id=group_id
            )

    @staticmethod
    def dump_servers(version=None, patterns="", persister=None):
        """Return the list of servers that are part of the groups
        listed in the patterns strings separated by comma.

        :param version: The connectors version of the data.
        :param patterns: group pattern.
        :param persister: Persister to persist the object to.
        :return: A list of servers belonging to the groups that match
                the provided pattern.
        """
        #This is used to store the consolidated list of servers
        #that will be returned.
        result_server_list = []

        #This stores the pattern that will be passed to the LIKE MySQL
        #command.
        like_pattern = None

        if patterns is None:
            patterns = ''

        #Split the patterns string into a list of patterns of groups.
        pattern_list = _utils.split_dump_pattern(patterns)

        #Iterate through the pattern list and fire a query for
        #each pattern.
        for find in pattern_list:
            if find == '':
                like_pattern = '%%'
            else:
                like_pattern = '%' + find + '%'
            rows = persister.exec_stmt(MySQLServer.DUMP_SERVERS,
                {"raw" : False, "params":(like_pattern, "FAULTY")})

            for row in rows:
                host, port = _server_utils.split_host_port(
                                row[2],
                                _server_utils.MYSQL_DEFAULT_PORT
                             )
                result_server_list.append(
                    (row[0], row[1], host, port, row[3], row[4],
                    row[5])
                )
        return _utils.wrap_output(result_server_list)

    @staticmethod
    def create(persister=None):
        """Create the objects(tables) that will store the Server
        information in the state store.

        :param persister: Persister to persist the object to.
        :raises: DatabaseError If the table already exists.
        """
        persister.exec_stmt(MySQLServer.CREATE_SERVER)

    @staticmethod
    def add_constraints(persister=None):
        """Add the constraints to the servers table.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(
                MySQLServer.ADD_FOREIGN_KEY_CONSTRAINT_GROUP_ID)
        return True

    @staticmethod
    def add(server, persister=None):
        """Write a server object into the state store.

        :param server: A reference to a server.
        :param persister: Persister to persist the object to.
        """
        assert(isinstance(server, MySQLServer))

        persister_uuid = persister.uuid
        if persister_uuid is not None and persister_uuid == server.uuid:
            raise _errors.ServerError("The MySQL Server instance used as Fabric's state store cannot be managed.")

        idx_mode = MySQLServer.get_mode_idx(server.mode)
        idx_status = MySQLServer.get_status_idx(server.status)
        persister.exec_stmt(MySQLServer.INSERT_SERVER,
            {"params":(str(server.uuid), server.address, idx_mode, idx_status,
            server.weight,
            server.group_id)}
        )

    def __eq__(self,  other):
        """Two servers are equal if they are both subclasses of MySQLServer
        and have equal UUID.
        """
        return isinstance(other, MySQLServer) and self.__uuid == other.uuid

    def __hash__(self):
        """A server is hashable through its uuid.
        """
        return hash(self.__uuid)

    def __str__(self):
        ret = "<server(uuid={0}, address={1}, mode={2}, status={3}>".\
            format(self.__uuid, self.__address, self.__mode, self.__status)
        return ret

def configure(config):
    """Set configuration values.
    """
    MySQLServer.USER = config.get("servers", "user")

    try:
        MySQLServer.PASSWD = config.get("servers", "password")
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        pass

    try:
        failover_interval = config.get("failure_tracking", "failover_interval")
        Group._FAILOVER_INTERVAL = int(failover_interval)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        pass



#
# Copyright (c) 2013 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Module for handling events in the Fabric.

Events represent something that happened either inside or outside the Fabric
node and events can trigger execution of procedures.

Central to the reception and processing of events is the **Event
Handler** (or just **Handler** when it is clear from the context),
which receives events from an external or internal source and enqueues
zero or more procedures for each code block that has been registered with
the event.

.. seqdiag::

   diagram {
     Source -> Handler [ label = "trigger(event)" ]
     Handler -> Handler [ label = "lookup(event)" ]
     Handler <-- Handler [ label = "blocks" ]
     Handler -> Executor [ label = "enqueue_procedures(block)",
                           note = "for block in blocks: \ \n
                           executor.enqueue_procedures(block)" ]
     Handler <-- Executor [ label = "procedure",
                            note = "procedures.append(procedure)" ]
     Source <-- Handler [ label = "procedure" ]
   }
"""
import functools
import logging

import mysql.fabric.errors as _errors
import mysql.fabric.executor as _executor

from mysql.fabric.utils import Singleton

_LOGGER = logging.getLogger(__name__)

def on_event(event):
    """Decorator to attach a callable to one event.

    The decorator also defined an ``undo`` decorator for the function,
    which allow abort actions to be executed if the body of the
    function throw an exception.

    Example use::

        @on_event(_events.SLAVE_PROMOTED)
        def changes_status(group_id, server_uuid):
            ....
        @changes_status.undo
        def changes_status_undo(group_id, server_uuid):
            ...

    The wrapped function will automatically call the ``undo`` callable
    if the main function raises an exception.
    """
    def register_func(func):
        """Wrapper that registers the function and attaches wrappers
        to the provided function."""
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            "Wrapper that execute undo function on an exception."
            try:
                _LOGGER.debug("Executing %s", wrapped.function.__name__)
                return wrapped.function(*args, **kwargs)
            except Exception:
                _LOGGER.debug("%s failed, executing compensation",
                              func.__name__)
                if wrapped.undo_function is not None:
                    wrapped.undo_function(*args, **kwargs)
                raise

        def undo_decorate(undo_func):
            "Undo decorator."
            wrapped.undo_function = undo_func

        wrapped.undo = undo_decorate
        wrapped.function = func
        wrapped.undo_function = None
        Handler().register(event, wrapped)
        return wrapped

    return register_func


class Event(object):
    """An event that can be triggered either from an external source or an
    internal source.

    An event might have a name, but it is optional. Only events with
    names can be triggered externally, but any event can be triggered
    internally.

    :param string name: The name of the event instance.

    For an example on how to trigger an event internally, see
    :meth:`Handler.trigger`.

    """
    def __init__(self, name=None):
        self.__name = name

    @property
    def name(self):
        """The name of the event.

        :returns: The name of the event, or None if it does not have a
                  name.
        """
        return self.__name


class Handler(Singleton):
    """An event handler to manage and trigger events in the system.

    The event handler is responsible for keeping track of all events
    and will also keep track of what code blocks should be executed
    when an event is triggered.

    """
    def __init__(self):
        """Constructor for Handler.
        """
        super(Handler, self).__init__()
        self.__executor = _executor.Executor()
        self.__instance_for = {}
        self.__blocks_for = {}

    def start(self):
        """Start the executor.
        """
        self.__executor.start()

    def shutdown(self):
        """Shut down the executor.
        """
        self.__executor.shutdown()

    def wait(self):
        """Wait until the executor stops.
        """
        self.__executor.wait()

    def register(self, event, blocks):
        """Register code blocks with an event in the event handler.

        This method register the code blocks supplied under the
        event. Each code block is represented as a callable (which
        means either a function or a class supporting the ``__call__``
        method).

        If the event was not previously registered, a new entry will
        be created for it. This involves both registering the name of
        the event (if it has a name) and the event instance.

        :param event: Event to register code blocks for.
        :param blocks: Callable to register for event.
        :type blocks: Callable or sequence of callables

        :except NotEventError: Trying to register code blocks with something
                               that is not an instance of :class:`Event`.
        :except NotCallableError: Trying to register something that is not a
                               callable with an event.
        """

        if not isinstance(event, Event):
            raise _errors.NotEventError(
                "Not possible to register with non-event")

        if callable(blocks):
            blocks = [blocks]

        # Check that all provided blocks are callables.
        try:
            for block in blocks:
                if not callable(block):
                    raise _errors.NotCallableError(
                        "Not possible to register non-callables")
        except TypeError:
            raise _errors.NotCallableError("Expected an iterable")

        _LOGGER.debug("Registering blocks %s for event %s under name %s",
                      blocks, event, event.name)

        # Register the name if not registered
        if event.name is not None and event.name not in self.__instance_for:
            self.__instance_for[event.name] = event

        # Register the callables
        self.__blocks_for.setdefault(event, set()).update(blocks)

    def unregister(self, event, block):
        """Unregister a code block from an event.

        :except NotEventError: Trying to unregister code blocks with something
                               that is not an instance of :class:`Event`.
        :except NotCallableError: Trying to unregister something that is not a
                                  callable.
        :except UnknownCallableError: The callable provided was not known to the
                                      event handler.
        """

        if not isinstance(event, Event):
            raise _errors.NotEventError(
                "Not possible to unregister with non-event")

        if not callable(block):
            raise _errors.NotCallableError(
                "Not possible to unregister a non-callable")

        _LOGGER.debug("Unregistering %s from event %s", block, event)

        try:
            self.__blocks_for[event].remove(block)
            if self.__blocks_for[event]:
                del self.__blocks_for[event]
        except KeyError:
            raise _errors.UnknownCallableError(
                "Not possible to unregister a non-existing block")

    def is_registered(self, event, block):
        """Check if a callable is registered with an event.

        Note that the exact callable instance is checked for, so you have to
        pass the instance that you want to check for.

        :param event: Event to check if callable is registered with.
        :param block: Callable to look for.
        :type block: callable

        :return: ``True`` if the callable instance is registered with the event,
                 ``False`` otherwise.
        :rtype: Boolean.
        """

        if not isinstance(event, Event):
            raise _errors.NotEventError(
                "Not possible to check registration for non-event")

        if not callable(block):
            raise _errors.NotCallableError(
                "Not possible to check for non-callable")

        try:
            return block in self.__blocks_for[event]
        except KeyError:
            return False

    def trigger(self, within_procedure, event,
                lockable_objects=None, *args, **kwargs):
        """Trigger an event.

        This function will trigger an event resulting in zero or more
        blocks being scheduled for execution by creating one
        :class:`Job` for each block. If any arguments are passed to
        this function they are passed to the :class:`Job` instance
        which will pass them to the code block on execution.

        Event can be triggered either by name or by instance. If a
        name is provided as *event*, then the event instance is looked
        up first and the instances triggered.

        If an instance is provided as event, it is looked up
        internally to find all blocks associated with it. The blocks
        are then scheduled (by creating one :class:`Job` for each
        block) with the executor and the function returns. This means
        that the blocks may not have been executed on return from the
        function.

        To be able to wait for the blocks to be completed before
        proceeding after triggering the event, references to the
        created procedures (containing the block) are returned. This means
        that if you want to trigger an event and wait for them to
        finish executing, you can use the following code::

           procedures = handler.trigger(False, SERVER_LOST, lockable_objects,
                                        "my.example.com")
           for procedure in procedures:
              procedure.wait()

        :within_procedure: Define if a new procedure will be created or not.
        :param event: Event to trigger.
        :type event: Event name or event instance
        :param lockable_objects: Set of objects to be locked by the concurrency
                                 control mechanism.
        :param args: Non-keyworded arguments to pass to the event.
        :param kwargs: Keyworded arguments to pass to the event.
        :returns: Procedures that were created as a result of triggering the
                  event.
        :rtype: Procedures that were scheduled.

        See :meth:`~mysql.fabric.executor.Executor.enqueue_procedures`.
        """
        _LOGGER.debug("Triggering event %s", event)

        if isinstance(event, basestring):
            if event in self.__instance_for:
                event = self.__instance_for[event]
            else:
                event = None

        # Enqueue the procedures and return a list with the procedures
        # scheduled.
        actions = [
            {"action" : (block, "Triggered by %s." % (event, ), args, kwargs),
             "job" : None
            }
            for block in self.__blocks_for.get(event, [])
        ]
        return self.__executor.enqueue_procedures(
            within_procedure, actions, lockable_objects
        )

def trigger(event, lockable_objects=None, *args, **kwargs):
    """Trigger an event by name or instance.

    :param event: The event to trigger.
    :type event: Event name or event instance.
    :param lockable_objects: Set of objects to be locked by the concurrency
                             control mechanism.
    :param args: Non-keyworded arguments to pass to the event.
    :param kwargs: Keyworded arguments to pass to the event.
    """
    handler = Handler()
    _LOGGER.debug("Triggering event %s in handler %s", event, handler)
    return handler.trigger(False, event, lockable_objects, *args, **kwargs)

def trigger_within_procedure(event, *args, **kwargs):
    """Trigger an event by name or instance. However, any job created
    due to this operation shall be created in the context of the current
    procedure.

    This method must be called within a job otherwise the
    :class:`~mysql.fabric.errors.ProgrammingError` exception will be raised.

    :param event: The event to trigger.
    :type event: Event name or event instance.
    :param args: Non-keyworded arguments to pass to the event.
    :param kwargs: Keyworded arguments to pass to the event.
    """
    handler = Handler()
    _LOGGER.debug("Triggering event %s in handler %s", event, handler)
    return handler.trigger(True, event, None, *args, **kwargs)

# Some pre-defined events. These are documented directly in the documentation
# and not using autodoc.
SERVER_LOST = Event("SERVER_LOST")
SERVER_PROMOTED = Event("SERVER_PROMOTED")
SERVER_DEMOTED = Event("SERVER_DEMOTED")



#
# Copyright (c) 2011, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This module contains classes and functions used to determine what MySQL
utilities are installed, their options, and usage. This module can be
used to allow a client to provide auto type and option completion.
"""

import glob
import os
import sys
import re
import subprocess

from mysql.utilities import AVAILABLE_UTILITIES
from mysql.utilities.common.format import print_dictionary_list
from mysql.utilities.common.tools import check_python_version
from mysql.utilities.exception import UtilError


_MAX_WIDTH = 78

# These utilities should not be used with the console
_EXCLUDE_UTILS = ['mysqluc', ]

RE_USAGE = (
    r"(?P<Version>.*?)"
    r"(?P<Usage>Usage:\s.*?)\w+\s\-\s"  # This match first
    # section <Usage> matching all till find a " - "
    r"(?P<Description>.*?)"  # Description is the text next
    # to " - " and till next match.
    r"(?P<O>\w*):"  # This is beginning of Options section
    r"(?P<Options>.*(?=^Introduction.\-{12})|.*$)"
    # match Options till end or till find Introduction -.
    r"(?:^Introduction.\-{12}){0,1}"  # not catching group
    r"(?P<Introduction>.*(?=^Helpful\sHints.\-{13})|.*$)"
    # captures Introduction (optional)
    # it will match Introduction till end or till Hints -
    r"(?:^Helpful\sHints.\-{13}){0,1}"  # Not catching group
    r"(?P<Helpful_Hints>.*)"
    # captures Helpful Hints (optional)
)

RE_OPTIONS = (
    r"^(?P<Alias>\s\s\-.*?)\s{2,}"  # Option Alias
    # followed by 2 o more spaces is his description
    r"(?P<Desc>.*?)(?=^\s\s\-)"  # description is all
    # text till not found other alias in the form
    # <-|--Alias> at the begin of the line.
)

RE_OPTION = r"\s+\-\-(.*?)\s"  # match Alias of the form <--Alias>

RE_ALIAS = r"\s+\-(\w+)\s*"  # match Alias of the form <-Alias>

WARNING_FAIL_TO_READ_OPTIONS = ("WARNING: {0} failed to read options."
                                " This utility will not be shown in 'help "
                                "utilities' and cannot be accessed from the "
                                "console.")


def get_util_path(default_path=''):
    """Find the path to the MySQL utilities

    This method will attempt to

    default_path[in]   provides known location of utilities
                       if provided, method will search this location first
                       before searching PYTHONPATH

    Returns string - path to utilities or None if not found
    """
    def _search_paths(needles, paths):
        """Search and return normalized path
        """
        for path in paths:
            norm_path = os.path.normpath(path)
            hay_stack = [os.path.join(norm_path, n) for n in needles]
            for needle in hay_stack:
                if os.path.isfile(needle):
                    return norm_path

        return None

    needle_name = 'mysqlreplicate'
    needles = [needle_name + ".py"]
    if os.name == "nt":
        needles.append(needle_name + ".exe")
    else:
        needles.append(needle_name)

    # Try the default by itself
    path_found = _search_paths(needles, [default_path])
    if path_found:
        return path_found

    # Try the pythonpath environment variable
    pythonpath = os.getenv("PYTHONPATH")
    if pythonpath:
        #This is needed on windows without a python setup, cause needs to
        #find the executable scripts.
        path = _search_paths(needles, [os.path.join(n, "../")
                                       for n in pythonpath.split(";", 1)])
        if path:
            return path
        path = _search_paths(needles, pythonpath.split(";", 1))
        if path:
            return path

    # Try the system paths
    path_found = _search_paths(needles, sys.path)
    if path_found:
        return path_found

    return None


class Utilities(object):
    """The utilities class can be used to discover what utilities are installed
    on the system as well as the usage and options for each utility.

    The list of utilities are read at initialization.

    This class is designed to support the following operations:

        get_util_matches()    - find all utilities that match a prefix
        get_option_matches()  - find all options that match a prefix for a
                                given utility
        get_usage()           - return the usage statement for a given utility
        show_utilities()      - display a 2-column list of utilities and their
                                descriptions
        show_options()        - display a 2-column list of the options for a
                                given utility including the name and
                                description of each option
    """

    def __init__(self, options=None):
        """Constructor
        """
        if options is None:
            options = {}
        self.util_list = []
        self.width = options.get('width', _MAX_WIDTH)
        self.util_path = get_util_path(options.get('utildir', ''))
        self.extra_utilities = options.get('add_util', {})
        self.hide_utils = options.get('hide_util', False)

        self.program_usage = re.compile(RE_USAGE, re.S | re.M)
        self.program_options = re.compile(RE_OPTIONS, re.S | re.M)
        self.program_option = re.compile(RE_OPTION)
        self.program_name = re.compile(RE_ALIAS)

        self.util_cmd_dict = {}
        self.posible_utilities = {}
        self.posible_utilities.update(AVAILABLE_UTILITIES)
        if self.extra_utilities and self.hide_utils:
            self.posible_utilities = self.extra_utilities
        else:
            self.posible_utilities.update(self.extra_utilities)
        self.available_utilities = self.posible_utilities
        for util_name, ver_compatibility in self.posible_utilities.iteritems():
            name_utility = "{0} utility".format(util_name)
            if ver_compatibility:
                min_v, max_v = ver_compatibility
                res = check_python_version(min_version=min_v,
                                           max_version=max_v,
                                           name=name_utility,
                                           print_on_fail=False,
                                           exit_on_fail=False,
                                           return_error_msg=True)
            else:
                res = check_python_version(name=name_utility,
                                           print_on_fail=False,
                                           exit_on_fail=False,
                                           return_error_msg=True)
            if isinstance(res, tuple):
                is_compat, error_msg = res
                if not is_compat:
                    self.available_utilities.remove(util_name)
                    print(WARNING_FAIL_TO_READ_OPTIONS.format(util_name))
                    print("ERROR: {0}\n".format(error_msg))
                    continue
            self._find_utility_cmd(util_name)

    @staticmethod
    def find_executable(util_name):
        """Search the system path for an executable matching the utility

        util_name[in]  Name of utility

        Returns string - name of executable (util_name or util_name.exe) or
                         original name if not found on the system path
        """
        paths = os.getenv("PATH").split(os.pathsep)
        for path in paths:
            new_path = os.path.join(path, util_name + "*")
            if os.name == "nt":
                new_path = '"{0}"'.format(new_path)
            found_path = glob.glob(new_path)
            if found_path:
                return os.path.split(found_path[0])[1]
        return util_name

    def _find_utility_cmd(self, utility_name):
        """ Locate the utility scripts

        util_name[in]   utility to find

        This method builds a dict of commands for invoke the utilities.
        """
        util_path = self.find_executable(os.path.join(self.util_path,
                                                      utility_name))
        util_path_parts = os.path.split(util_path)
        parts = os.path.splitext(util_path_parts[len(util_path_parts) - 1])
        # filter extensions
        exts = ['.py', '.exe', '', 'pyc']
        if (parts[0] not in _EXCLUDE_UTILS and
           (len(parts) == 1 or (len(parts) == 2 and parts[1] in exts))):
            util_name = str(parts[0])
            file_ext = parts[1]
            command = "{0}{1}".format(util_name, file_ext)

            util_path = self.util_path
            utility_path = command
            if not os.path.exists(command):
                utility_path = os.path.join(util_path, utility_name)

            # Now try the extensions
            if not os.path.exists(utility_path):
                if file_ext:
                    utility_path = "{0}{1}".format(utility_path, file_ext)
                else:
                    for ext in exts:
                        try_path = "{0}{1}".format(utility_path, ext)
                        if os.path.exists(try_path):
                            utility_path = try_path

            if not os.path.exists(utility_path):
                print("WARNING: Unable to locate utility {0}."
                      "".format(utility_name))
                print(WARNING_FAIL_TO_READ_OPTIONS.format(util_name))
                return

            # Check for running against .exe
            if utility_path.endswith(".exe"):
                cmd = []
            # Not using .exe
            else:
                cmd = [sys.executable]

            cmd.extend([utility_path])
            self.util_cmd_dict[utility_name] = tuple(cmd)

    def find_utilities(self, this_utils=None):
        """ Locate the utility scripts
        this_utils[in]   list of utilities to find, default None to find all.

        This method builds a list of utilities.
        """

        if not this_utils:
            # Not utilities name to find was passed, find help for all those
            # utilities not previously found in a previos call.
            utils = self.available_utilities
            working_utils = [util['name'] for util in self.util_list]
            if not len(working_utils) < len(self.util_list):
                utils = [name for name in utils if not name in working_utils]
            if len(utils) < 1:
                return
        else:
            # utilities name given to find for, find help for all these which
            # was not previously found in a previos call.
            working_utils = [util['name'] for util in self.util_list]
            utils = [util for util in this_utils if util not in working_utils]
            if len(utils) < 1:
                return

        # Execute the utility command using get_util_info()
        # that returns --help partially parsed.
        for util_name in utils:
            if util_name in self.util_cmd_dict:
                cmd = self.util_cmd_dict.pop(util_name)
                util_info = self.get_util_info(list(cmd), util_name)
                if util_info and util_info["usage"]:
                    util_info["cmd"] = tuple(cmd)
                    self.util_list.append(util_info)
                    working_utils.append(util_name)

        self.util_list.sort(key=lambda util_list: util_list['name'])

    def get_util_info(self, cmd, util_name):
        """Get information about utility

        cmd[in]        a list with the elements that conform the command
                       to invoke the utility
        util_name[in]  name of utility to get information

        Returns dictionary - name, description, usage, options
        """
        cmd.extend(["--help"])
        # rmv print('executing ==> {0}'.format(cmd))
        try:
            proc = subprocess.Popen(cmd, shell=False,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            stdout_temp, stderr_temp = proc.communicate()
            returncode = proc.returncode
        except OSError:
            # always OS error if not found.
            # No such file or directory
            stdout_temp = ""
            returncode = 0

        # Parse the help output and save the information found
        usage = None
        description = None

        if stderr_temp or returncode:
            print(WARNING_FAIL_TO_READ_OPTIONS.format(util_name))
            if stderr_temp:
                print("The execution of the command returned: {0}"
                      "".format(stderr_temp))
            else:
                print("UNKNOWN. To diagnose, exit mysqluc and attempt the "
                      "command: {0} --help".format(util_name))
            return None

        res = self.program_usage.match(stdout_temp.replace("\r", ""))
        if not res:
            print(WARNING_FAIL_TO_READ_OPTIONS.format(util_name))
            print("An error occurred while trying to parse the options "
                  "from the utility")
            return None
        else:
            usage = res.group("Usage").replace("\n", "")
            desc_clean = res.group("Description").replace("\n", " ").split()
            description = (" ".join(desc_clean)) + " "
            # standardize string.
            Options = res.group("Options") + "\n  -"

        # Create dictionary for the information
        utility_data = {
            'name': util_name,
            'description': description,
            'usage': usage,
            'options': Options
        }
        return utility_data

    def parse_all_options(self, utility):
        """ Parses all options for the given utility.

        utility[inout]   that contains the options info to parse
        """
        options_info = utility['options']
        if isinstance(options_info, list):
            # nothing to do if it is a list.
            return

        options = []
        res = self.program_options.findall(options_info)

        for opt in res:
            option = {}
            name = self.program_option.search(opt[0] + " ")
            if name:
                option['name'] = str(name.group(1))
            alias = self.program_name.search(opt[0] + " ")
            if alias:
                option['alias'] = str(alias.group(1))
            else:
                option['alias'] = None

            desc_clean = opt[1].replace("\n", " ").split()
            option['description'] = " ".join(desc_clean)
            option['long_name'] = option['name']
            parts = option['name'].split('=')
            option['req_value'] = len(parts) == 2
            if option['req_value']:
                option['name'] = parts[0]
            if option:
                options.append(option)

        utility['options'] = options

    def get_util_matches(self, util_prefix):
        """Get list of utilities that match a prefix

        util_prefix[in] prefix for name of utility

        Returns dictionary entry for utility based on matching first n chars
        """
        matches = []
        if not util_prefix.lower().startswith('mysql'):
            util_prefix = 'mysql' + util_prefix
        for util in self.available_utilities:
            if util[0:len(util_prefix)].lower() == util_prefix:
                matches.append(util)
        # make sure the utilities description has been found for the matches.
        self.find_utilities(matches)
        matches = [util for util in self.util_list if util['name'] in matches]
        return matches

    def get_option_matches(self, util_info, option_prefix, find_alias=False):
        """Get list of option dictionary entries for options that match
        the prefix.

        util_info[in]     utility information
        option_prefix[in] prefix for option name
        find_alias[in]    if True, match alias (default = False)

        Returns list of dictionary items that match prefix
        """
        # Check type of util_info
        if util_info is None or util_info == {} or \
                not isinstance(util_info, dict):
            raise UtilError("Empty or invalide utility dictionary.")

        matches = []

        stop = len(option_prefix)
        if isinstance(util_info['options'], str):
            self.parse_all_options(util_info)
        for option in util_info['options']:
            if option is None:
                continue
            name = option.get('name', None)
            if name is None:
                continue
            if find_alias:
                if option.get('alias', '') == option_prefix:
                    matches.append(option)
            else:
                if name[0:stop] == option_prefix:
                    matches.append(option)

        return matches

    def show_utilities(self, print_list=None):
        """Show list of utilities as a 2-column list.

        print_list[in]    list of utilities to print - default is None
                          which means print all utilities
        """

        if print_list is None:
            if len(self.util_list) != len(self.available_utilities):
                self.find_utilities()
            list_of_utilities = self.util_list
        else:
            list_of_utilities = print_list
        print
        if len(list_of_utilities) > 0:
            print_dictionary_list(['Utility', 'Description'],
                                  ['name', 'description'],
                                  list_of_utilities, self.width)
        else:
            print
            print "No utilities match the search term."
        print

    def get_options_dictionary(self, utility_options):
        """Retrieve the options dictionary.

        This method builds a new dictionary that contains the options for the
        utilities read.

        utility_options[in]   list of options for utilities or the utility.

        Return dictionary - list of options for all utilities.
        """
        dictionary_list = []

        if isinstance(utility_options, dict):
            if isinstance(utility_options['options'], str):
                # options had not been parsed yet
                self.parse_all_options(utility_options)
            options = utility_options['options']
        else:
            options = utility_options

        for option in options:
            name = option.get('long_name', '')
            if len(name) == 0:
                continue
            name = '--' + name
            alias = option.get('alias', None)
            if alias is not None:
                name = '-' + alias + ", " + name
            item = {
                'long_name': name,
                'description': option.get('description', '')
            }
            dictionary_list.append(item)

        return dictionary_list

    def show_options(self, options):
        """Show list of options for a utility by name.

        options[in]    structure containing the options

        This method displays a list of the options and their descriptions
        for the given utility.
        """
        if len(options) > 0:
            dictionary_list = self.get_options_dictionary(options)
            print
            print
            print_dictionary_list(['Option', 'Description'],
                                  ['long_name', 'description'],
                                  dictionary_list, self.width)
            print

    @staticmethod
    def get_usage(util_info):
        """Get the usage statement for the utility

        util_info[in]  dictionary entry for utility information

        Returns string usage statement
        """
        # Check type of util_info
        if util_info is None or util_info == {} or \
                not isinstance(util_info, dict):
            return False

        return util_info['usage']
